from acryl_datahub_cloud.sdk.assertion import (
    SmartFreshnessAssertion,
    SmartVolumeAssertion,
)
from acryl_datahub_cloud.sdk.assertion_input import (
    AssertionIncidentBehavior,
    DetectionMechanism,
    FixedRangeExclusionWindow,
    InferenceSensitivity,
)
from acryl_datahub_cloud.sdk.assertions_client import AssertionsClient
from acryl_datahub_cloud.sdk.resolver_client import ResolverClient
from acryl_datahub_cloud.sdk.subscription_client import SubscriptionClient

__all__ = [
    "SmartFreshnessAssertion",
    "SmartVolumeAssertion",
    "DetectionMechanism",
    "InferenceSensitivity",
    "FixedRangeExclusionWindow",
    "AssertionIncidentBehavior",
    "AssertionsClient",
    "ResolverClient",
    "SubscriptionClient",
]



from typing import List, Optional

from datahub.metadata.urns import (
    SubscriptionUrn,
)
from datahub.sdk.main_client import DataHubClient
from datahub.sdk.resolver_client import ResolverClient as OSSResolverClient
from datahub.sdk.search_filters import Filter, FilterDsl as F


class ResolverClient(OSSResolverClient):
    def __init__(self, client: DataHubClient):
        super().__init__(client=client)

    def subscription(
        self,
        *,
        entity_urn: Optional[str] = None,
        actor_urn: Optional[str] = None,
    ) -> List[SubscriptionUrn]:
        """Retrieve subscriptions for a given entity or actor, or both if both are given.
        Args:
            entity_urn: Optional URN of the entity to filter subscriptions by.
            actor_urn: Optional URN of the actor to filter subscriptions by.
        Returns:
            List[SubscriptionUrn]: List of SubscriptionUrns matching the filters,
            empty list if none.
        """
        filters: List[Filter] = [F.entity_type(SubscriptionUrn.ENTITY_TYPE)]

        if entity_urn is not None:
            filters.append(F.custom_filter("entityUrn", "EQUAL", [entity_urn]))

        if actor_urn is not None:
            filters.append(F.custom_filter("actorUrn", "EQUAL", [actor_urn]))

        filter = F.and_(*filters)
        subscriptions = list(self._client.search.get_urns(filter=filter))
        return [SubscriptionUrn.from_string(urn) for urn in subscriptions]



import json
import time
from typing import Any, Dict, List, cast
from unittest.mock import MagicMock

import pytest

from acryl_datahub_cloud.datahub_forms_notifications.forms_notifications_source import (
    DataHubFormsNotificationsSource,
    DataHubFormsNotificationsSourceConfig,
)
from acryl_datahub_cloud.datahub_forms_notifications.query import (
    GRAPHQL_GET_FEATURE_FLAG,
    GRAPHQL_GET_SEARCH_RESULTS_TOTAL,
    GRAPHQL_SEND_FORM_NOTIFICATION_REQUEST,
)
from acryl_datahub_cloud.notifications.notification_recipient_builder import (
    NotificationRecipientBuilder,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.schema_classes import (
    CorpUserAppearanceSettingsClass,
    CorpUserSettingsClass,
    FormActorAssignmentClass,
    FormInfoClass,
    FormNotificationDetailsClass,
    FormNotificationEntryClass,
    FormNotificationsClass,
    FormStateClass,
    FormStatusClass,
    FormTypeClass,
    NotificationSettingClass,
    NotificationSettingsClass,
    NotificationSettingValueClass,
    NotificationSinkTypeClass,
    SlackNotificationSettingsClass,
)


@pytest.fixture
def mock_graph() -> MagicMock:
    return MagicMock(spec=DataHubGraph)


@pytest.fixture
def mock_context(mock_graph: MagicMock) -> MagicMock:
    ctx = MagicMock(spec=PipelineContext)
    ctx.require_graph.return_value = mock_graph
    return ctx


@pytest.fixture
def basic_config() -> DataHubFormsNotificationsSourceConfig:
    return DataHubFormsNotificationsSourceConfig(form_urns=["urn:li:form:test-form-1"])


@pytest.fixture
def source(
    mock_context: MagicMock, basic_config: DataHubFormsNotificationsSourceConfig
) -> DataHubFormsNotificationsSource:
    source = DataHubFormsNotificationsSource(basic_config, mock_context)
    # Initialize the graph client and ensure it's typed as a MagicMock
    source.graph = cast(DataHubGraph, mock_context.require_graph())
    assert source.graph is not None  # For mypy
    return source


def test_init(
    source: DataHubFormsNotificationsSource,
    mock_context: MagicMock,
    basic_config: DataHubFormsNotificationsSourceConfig,
) -> None:
    """Test basic initialization of the source"""
    assert source.config == basic_config
    assert source.ctx == mock_context
    assert source.report is not None
    assert source.report.notifications_sent == 0


def test_execute_graphql_with_retry_success(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test successful GraphQL execution"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    expected_response = {"data": {"some": "data"}}
    mock_graph.execute_graphql.return_value = expected_response

    response = source.execute_graphql_with_retry("query {}", {})

    assert response == expected_response
    mock_graph.execute_graphql.assert_called_once_with("query {}", variables={})


def test_execute_graphql_with_retry_error(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test GraphQL execution with error"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    mock_graph.execute_graphql.return_value = {"error": "Something went wrong"}

    with pytest.raises(Exception, match="GraphQL error: Something went wrong"):
        source.execute_graphql_with_retry("query {}", {})


def test_get_forms_with_config_urns(source: DataHubFormsNotificationsSource) -> None:
    """Test getting forms from config URNs"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(users=["urn:li:corpuser:test-user"]),
    )

    mock_graph.get_entities.return_value = {
        "urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)}
    }

    forms = source.get_forms()
    assert len(forms) == 1
    assert forms[0][0] == "urn:li:form:test-form-1"
    assert forms[0][1] == mock_form_info


def test_get_forms_with_search(source: DataHubFormsNotificationsSource) -> None:
    """Test getting forms through search when no URNs provided"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Override config to have no form URNs
    source.config.form_urns = None

    # Mock the search response
    mock_graph.execute_graphql.return_value = {
        "scrollAcrossEntities": {
            "nextScrollId": None,
            "searchResults": [
                {"entity": {"urn": "urn:li:form:test-form-1"}},
                {"entity": {"urn": "urn:li:form:test-form-2"}},
            ],
        }
    }

    # Mock the form info response
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(users=["urn:li:corpuser:test-user"]),
    )

    mock_graph.get_entities.return_value = {
        "urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)},
        "urn:li:form:test-form-2": {"formInfo": (mock_form_info, None)},
    }

    forms = source.get_forms()
    assert len(forms) == 2
    assert forms[0][0] == "urn:li:form:test-form-1"
    assert forms[1][0] == "urn:li:form:test-form-2"


def test_get_form_assignees(source: DataHubFormsNotificationsSource) -> None:
    """Test getting form assignees including users and group members"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    form = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(
            users=["urn:li:corpuser:user1"], groups=["urn:li:corpGroup:group1"]
        ),
    )

    # Mock group members
    mock_graph.get_related_entities.return_value = [
        MagicMock(urn="urn:li:corpuser:group-user1"),
        MagicMock(urn="urn:li:corpuser:group-user2"),
    ]

    assignees = source.get_form_assignees("urn:li:form:test-form-1", form)
    assert len(assignees) == 3
    assert "urn:li:corpuser:user1" in assignees
    assert "urn:li:corpuser:group-user1" in assignees
    assert "urn:li:corpuser:group-user2" in assignees


def test_is_form_complete(source: DataHubFormsNotificationsSource) -> None:
    """Test checking if a form is complete"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Test complete form
    mock_graph.execute_graphql.return_value = {"searchAcrossEntities": {"total": 0}}

    assert (
        source.is_form_complete("urn:li:form:test-form-1", FormTypeClass.COMPLETION)
        is True
    )

    # Test incomplete form
    mock_graph.execute_graphql.return_value = {"searchAcrossEntities": {"total": 5}}

    assert (
        source.is_form_complete("urn:li:form:test-form-1", FormTypeClass.COMPLETION)
        is False
    )


def test_process_notify_on_publish(source: DataHubFormsNotificationsSource) -> None:
    """Test processing notifications for form publish"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock the recipient builder
    mock_recipient_builder = MagicMock(spec=NotificationRecipientBuilder)
    mock_recipient_builder.build_actor_recipients.return_value = [
        {
            "actor": "urn:li:corpuser:user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        }
    ]
    mock_recipient_builder.convert_recipients_to_json_objects.return_value = [
        {
            "actor": "urn:li:corpuser:user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        }
    ]
    source.recipient_builder = mock_recipient_builder

    # Mock successful notification send
    mock_graph.execute_graphql.return_value = {"sendFormNotificationRequest": True}

    source.process_notify_on_publish(
        ["urn:li:corpuser:user1"], "Test Form", "urn:li:form:test-form-1"
    )

    # Verify the notification was sent with correct parameters
    mock_graph.execute_graphql.assert_called_once()
    call_args = mock_graph.execute_graphql.call_args[1]
    variables = call_args["variables"]

    # Verify notification type and parameters
    assert variables["input"]["type"] == "BROADCAST_COMPLIANCE_FORM_PUBLISH"
    assert variables["input"]["parameters"][0]["value"] == "Test Form"

    # Verify recipients structure
    recipients = variables["input"]["recipients"]
    assert isinstance(recipients, list)
    assert len(recipients) == 1  # One recipient for one user

    # Verify recipient structure
    recipient = recipients[0]
    assert "actor" in recipient
    assert recipient["actor"] == "urn:li:corpuser:user1"
    assert "type" in recipient
    assert recipient["type"] == "COMPLIANCE_FORM_PUBLISH"
    assert "parameters" in recipient
    assert isinstance(recipient["parameters"], list)

    # Verify the recipient builder was called correctly
    mock_recipient_builder.build_actor_recipients.assert_called_once_with(
        ["urn:li:corpuser:user1"], "COMPLIANCE_FORM_PUBLISH", True
    )


def test_get_workunits(source: DataHubFormsNotificationsSource) -> None:
    """Test getting workunits from the source"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock form info
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(users=["urn:li:corpuser:test-user"]),
    )

    mock_graph.get_entities.return_value = {
        "urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)}
    }

    # Mock form completion check
    mock_graph.execute_graphql.side_effect = [
        # First call for form info
        {"data": {"some": "data"}},
        # Second call for is_form_complete
        {"searchAcrossEntities": {"total": 0}},
    ]

    workunits = list(source.get_workunits())
    assert len(workunits) == 0  # This source doesn't produce work units


def test_get_incomplete_assets_for_form(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test getting the correct filter based on form type"""
    # Test completion form filter
    completion_filter = source._get_incomplete_assets_for_form(
        "urn:li:form:test-form-1", FormTypeClass.COMPLETION
    )
    assert len(completion_filter) == 1
    assert completion_filter[0]["and"][0]["field"] == "incompleteForms"
    assert completion_filter[0]["and"][0]["values"] == ["urn:li:form:test-form-1"]

    # Test verification form filter
    verification_filter = source._get_incomplete_assets_for_form(
        "urn:li:form:test-form-1", FormTypeClass.VERIFICATION
    )
    assert len(verification_filter) == 2

    # Check first condition (incomplete forms)
    assert verification_filter[0]["and"][0]["field"] == "incompleteForms"
    assert verification_filter[0]["and"][0]["values"] == ["urn:li:form:test-form-1"]

    # Check second condition (completed but not verified)
    assert verification_filter[1]["and"][0]["field"] == "completedForms"
    assert verification_filter[1]["and"][0]["values"] == ["urn:li:form:test-form-1"]
    assert verification_filter[1]["and"][1]["field"] == "verifiedForms"
    assert verification_filter[1]["and"][1]["values"] == ["urn:li:form:test-form-1"]
    assert verification_filter[1]["and"][1]["negated"] is True


def test_notify_form_assignees(source: DataHubFormsNotificationsSource) -> None:
    """Test the complete flow of notifying form assignees"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock the recipient builder
    mock_recipient_builder = MagicMock(spec=NotificationRecipientBuilder)
    mock_recipient_builder.build_actor_recipients.return_value = [
        {
            "actor": "urn:li:corpuser:user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        },
        {
            "actor": "urn:li:corpuser:group-user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        },
    ]
    mock_recipient_builder.convert_recipients_to_json_objects.return_value = [
        {
            "actor": "urn:li:corpuser:user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        },
        {
            "actor": "urn:li:corpuser:group-user1",
            "type": "COMPLIANCE_FORM_PUBLISH",
            "parameters": [],
        },
    ]
    source.recipient_builder = mock_recipient_builder

    # Mock form info
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(
            users=["urn:li:corpuser:user1"], groups=["urn:li:corpGroup:group1"]
        ),
    )

    # Mock form notifications (empty for all users)
    mock_form_notifications = FormNotificationsClass(notificationDetails=[])

    # Mock get_entities to return different responses based on the call
    def mock_get_entities(
        entity_type: str, urns: List[str], aspects: Any
    ) -> Dict[str, Any]:
        if entity_type == "form":
            return {"urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)}}
        elif entity_type == "corpuser":
            return {
                urn: {"formNotifications": (mock_form_notifications, None)}
                for urn in urns
            }
        return {}

    mock_graph.get_entities.side_effect = mock_get_entities

    # Mock group members
    mock_graph.get_related_entities.return_value = [
        MagicMock(urn="urn:li:corpuser:group-user1"),
        MagicMock(urn="urn:li:corpuser:group-user2"),
    ]

    # Mock form completion check
    mock_graph.execute_graphql.side_effect = [
        # First call for is_form_complete
        {"searchAcrossEntities": {"total": 5}},  # Form is incomplete
        # Second call for sending notification
        {"sendFormNotificationRequest": True},
    ]

    # Execute the notification flow
    source.notify_form_assignees()

    # Verify the form was retrieved
    mock_graph.get_entities.assert_any_call(
        "form", ["urn:li:form:test-form-1"], ["formInfo"]
    )

    # Verify form notifications were retrieved for all users
    # Get all calls to get_entities
    get_entities_calls = [
        call
        for call in mock_graph.get_entities.call_args_list
        if call[0][0] == "corpuser"  # Filter for corpuser calls
    ]
    assert len(get_entities_calls) == 1  # Should be exactly one call for corpuser

    # Get the actual URNs from the call
    actual_urns = get_entities_calls[0][0][1]  # Get the URNs list from the call
    expected_urns = [
        "urn:li:corpuser:user1",
        "urn:li:corpuser:group-user1",
        "urn:li:corpuser:group-user2",
    ]

    # Compare sorted lists
    assert sorted(actual_urns) == sorted(expected_urns)
    assert get_entities_calls[0][0][2] == ["formNotifications"]  # Verify aspects

    # Verify form completion was checked
    mock_graph.execute_graphql.assert_any_call(
        GRAPHQL_GET_SEARCH_RESULTS_TOTAL,
        variables={
            "count": 0,
            "orFilters": source._get_incomplete_assets_for_form(
                "urn:li:form:test-form-1", FormTypeClass.COMPLETION
            ),
        },
    )

    # Verify group members were fetched
    mock_graph.get_related_entities.assert_called_once_with(
        "urn:li:corpGroup:group1",
        ["IsMemberOfGroup", "IsMemberOfNativeGroup"],
        mock_graph.RelationshipDirection.INCOMING,
    )

    # Verify notification was sent
    mock_graph.execute_graphql.assert_any_call(
        GRAPHQL_SEND_FORM_NOTIFICATION_REQUEST,
        variables={
            "input": {
                "type": "BROADCAST_COMPLIANCE_FORM_PUBLISH",
                "parameters": [{"key": "formName", "value": "Test Form"}],
                "recipients": mock_recipient_builder.build_actor_recipients.return_value,
            }
        },
    )

    # Verify recipient builder was called with all assignees
    expected_assignees = [
        "urn:li:corpuser:user1",
        "urn:li:corpuser:group-user1",
        "urn:li:corpuser:group-user2",
    ]
    actual_call = mock_recipient_builder.build_actor_recipients.call_args[0]
    assert sorted(actual_call[0]) == sorted(
        expected_assignees
    )  # Sort both lists before comparison
    assert actual_call[1] == "COMPLIANCE_FORM_PUBLISH"
    assert actual_call[2] is True

    # Verify report counters
    assert source.report.notifications_sent == 2  # Two recipients in the mock response
    assert source.report.forms_count == 1  # One form processed


def test_notify_form_assignees_no_recipients(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test that report counters are not updated when there are no recipients"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock the recipient builder to return no recipients
    mock_recipient_builder = MagicMock(spec=NotificationRecipientBuilder)
    mock_recipient_builder.build_actor_recipients.return_value = []
    source.recipient_builder = mock_recipient_builder

    # Mock form info
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(users=[]),  # No users or groups
    )

    # Mock get_forms response
    mock_graph.get_entities.return_value = {
        "urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)}
    }

    # Mock form completion check
    mock_graph.execute_graphql.return_value = {
        "searchAcrossEntities": {"total": 5}  # Form is incomplete
    }

    # Execute the notification flow
    source.notify_form_assignees()

    # Verify report counters are not updated
    assert source.report.notifications_sent == 0
    assert source.report.forms_count == 0


def test_get_form_assignees_with_owners(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test getting form assignees including owners from assets"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock form info with owners flag set to True
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(
            users=["urn:li:corpuser:user1"],
            groups=["urn:li:corpGroup:group1"],
            owners=True,  # Enable owner-based assignees
        ),
    )

    # Mock group members
    mock_graph.get_related_entities.return_value = [
        MagicMock(urn="urn:li:corpuser:group-user1"),
        MagicMock(urn="urn:li:corpuser:group-user2"),
    ]

    # Mock asset search results with owners
    mock_graph.get_results_by_filter.return_value = [
        {
            "extraProperties": [
                {"name": "urn", "value": json.dumps("urn:li:dataset:asset1")},
                {
                    "name": "owners",
                    "value": json.dumps(
                        ["urn:li:corpuser:owner1", "urn:li:corpGroup:owner-group1"]
                    ),
                },
            ]
        },
        {
            "extraProperties": [
                {"name": "urn", "value": json.dumps("urn:li:dataset:asset2")},
                {
                    "name": "owners",
                    "value": json.dumps(
                        ["urn:li:corpuser:owner2", "urn:li:corpuser:owner3"]
                    ),
                },
            ]
        },
    ]

    # Get assignees
    assignees = source.get_form_assignees("urn:li:form:test-form-1", mock_form_info)

    # Verify all types of assignees are included
    expected_assignees = [
        # Direct users from form
        "urn:li:corpuser:user1",
        # Group members
        "urn:li:corpuser:group-user1",
        "urn:li:corpuser:group-user2",
        # Asset owners
        "urn:li:corpuser:owner1",
        "urn:li:corpuser:owner2",
        "urn:li:corpuser:owner3",
    ]

    # Sort both lists for comparison since order doesn't matter
    assert sorted(assignees) == sorted(expected_assignees)

    # Verify the correct filter was used to get assets
    mock_graph.get_results_by_filter.assert_called_once()
    call_args = mock_graph.get_results_by_filter.call_args[1]
    assert call_args["extra_or_filters"] == source._get_incomplete_assets_for_form(
        "urn:li:form:test-form-1", FormTypeClass.COMPLETION
    )
    assert call_args["skip_cache"] is True


def test_filter_assignees_to_notify(source: DataHubFormsNotificationsSource) -> None:
    """Test filtering assignees who haven't been notified yet"""
    mock_graph = cast(MagicMock, source.graph)

    # Mock user settings with notifications enabled
    mock_settings = CorpUserSettingsClass(
        appearance=CorpUserAppearanceSettingsClass(),
        notificationSettings=NotificationSettingsClass(
            sinkTypes=[NotificationSinkTypeClass.SLACK],
            settings={
                "COMPLIANCE_FORM_PUBLISH": NotificationSettingClass(
                    value=NotificationSettingValueClass.ENABLED,
                    params={"slack.enabled": "true"},
                )
            },
            slackSettings=SlackNotificationSettingsClass(userHandle="test-user"),
        ),
    )

    # Mock form notifications for one user
    mock_form_notifications = FormNotificationsClass(
        notificationDetails=[
            FormNotificationDetailsClass(
                formUrn="urn:li:form:test-form",
                notificationLog=[
                    FormNotificationEntryClass(
                        time=int(time.time() * 1000),
                        notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
                    )
                ],
            )
        ]
    )

    mock_graph.get_entities.return_value = {
        "urn:li:corpuser:test-user": {
            "corpUserSettings": (mock_settings, None),
            "formNotifications": (mock_form_notifications, None),
        },
        "urn:li:corpuser:test-user-2": {
            "corpUserSettings": (mock_settings, None),
            "formNotifications": (None, None),  # No notifications yet
        },
    }

    # Test filtering users for a form they haven't been notified about
    filtered_users = source.filter_assignees_to_notify(
        ["urn:li:corpuser:test-user", "urn:li:corpuser:test-user-2"],
        "urn:li:form:different-form",
    )

    assert (
        len(filtered_users) == 2
    )  # Both users should be included for a different form

    # Test filtering users for a form they have been notified about
    filtered_users = source.filter_assignees_to_notify(
        ["urn:li:corpuser:test-user", "urn:li:corpuser:test-user-2"],
        "urn:li:form:test-form",
    )

    assert len(filtered_users) == 1  # Only test-user-2 should be included
    assert filtered_users[0] == "urn:li:corpuser:test-user-2"


def test_has_user_been_sent_notification(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test checking if a user has been sent a notification"""
    mock_form_notifications = FormNotificationsClass(
        notificationDetails=[
            FormNotificationDetailsClass(
                formUrn="urn:li:form:test-form",
                notificationLog=[
                    FormNotificationEntryClass(
                        time=int(time.time() * 1000),
                        notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
                    )
                ],
            )
        ]
    )

    # Test user who has been notified
    assert (
        source.has_user_been_sent_notification(
            "urn:li:corpuser:test-user",
            "urn:li:form:test-form",
            mock_form_notifications,
            "BROADCAST_COMPLIANCE_FORM_PUBLISH",
        )
        is True
    )

    # Test user who hasn't been notified for this form
    assert (
        source.has_user_been_sent_notification(
            "urn:li:corpuser:test-user",
            "urn:li:form:different-form",
            mock_form_notifications,
            "BROADCAST_COMPLIANCE_FORM_PUBLISH",
        )
        is False
    )

    # Test user who hasn't been notified for this notification type
    assert (
        source.has_user_been_sent_notification(
            "urn:li:corpuser:test-user",
            "urn:li:form:test-form",
            mock_form_notifications,
            "DIFFERENT_NOTIFICATION_TYPE",
        )
        is False
    )


def test_get_notification_details_for_form(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test getting notification details for a specific form"""
    mock_form_notifications = FormNotificationsClass(
        notificationDetails=[
            FormNotificationDetailsClass(
                formUrn="urn:li:form:test-form",
                notificationLog=[
                    FormNotificationEntryClass(
                        time=int(time.time() * 1000),
                        notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
                    )
                ],
            )
        ]
    )

    # Test getting details for existing form
    details = source.get_notification_details_for_form(
        "urn:li:corpuser:test-user",
        "urn:li:form:test-form",
        mock_form_notifications,
    )
    assert details is not None
    assert details.formUrn == "urn:li:form:test-form"

    # Test getting details for non-existent form
    details = source.get_notification_details_for_form(
        "urn:li:corpuser:test-user",
        "urn:li:form:different-form",
        mock_form_notifications,
    )
    assert details is None


def test_populate_user_to_form_notifications(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test populating user form notifications map"""
    mock_graph = cast(MagicMock, source.graph)

    mock_form_notifications = FormNotificationsClass(
        notificationDetails=[
            FormNotificationDetailsClass(
                formUrn="urn:li:form:test-form",
                notificationLog=[
                    FormNotificationEntryClass(
                        time=int(time.time() * 1000),
                        notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
                    )
                ],
            )
        ]
    )

    mock_graph.get_entities.return_value = {
        "urn:li:corpuser:test-user": {
            "formNotifications": (mock_form_notifications, None),
        },
        "urn:li:corpuser:test-user-2": {
            "formNotifications": (None, None),
        },
    }

    # Test populating map for new users
    source.populate_user_to_form_notifications(
        ["urn:li:corpuser:test-user", "urn:li:corpuser:test-user-2"]
    )

    assert "urn:li:corpuser:test-user" in source.user_to_form_notifications
    assert (
        source.user_to_form_notifications["urn:li:corpuser:test-user"]
        == mock_form_notifications
    )
    assert "urn:li:corpuser:test-user-2" not in source.user_to_form_notifications

    # Test that existing users aren't re-fetched
    mock_graph.get_entities.reset_mock()
    source.populate_user_to_form_notifications(["urn:li:corpuser:test-user"])
    mock_graph.get_entities.assert_not_called()


def test_update_form_notifications(source: DataHubFormsNotificationsSource) -> None:
    """Test updating form notifications for a user"""
    mock_graph = cast(MagicMock, source.graph)

    # Initial form notifications
    initial_notifications = FormNotificationsClass(
        notificationDetails=[
            FormNotificationDetailsClass(
                formUrn="urn:li:form:existing-form",
                notificationLog=[
                    FormNotificationEntryClass(
                        time=int(time.time() * 1000),
                        notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
                    )
                ],
            )
        ]
    )

    source.user_to_form_notifications["urn:li:corpuser:test-user"] = (
        initial_notifications
    )

    # Update notifications for a new form
    source.update_form_notifications(
        "urn:li:corpuser:test-user",
        "urn:li:form:new-form",
    )

    # Verify that the graph client was called to emit the updated notifications
    mock_graph.emit.assert_called_once()
    emitted_wrapper = mock_graph.emit.call_args[0][0]
    assert emitted_wrapper.entityUrn == "urn:li:corpuser:test-user"
    assert isinstance(emitted_wrapper.aspect, FormNotificationsClass)

    # Verify the updated notifications contain both forms
    updated_notifications = emitted_wrapper.aspect
    assert len(updated_notifications.notificationDetails) == 2

    # Verify the new form notification was added
    new_form_details = next(
        d
        for d in updated_notifications.notificationDetails
        if d.formUrn == "urn:li:form:new-form"
    )
    assert len(new_form_details.notificationLog) == 1
    assert (
        new_form_details.notificationLog[0].notificationType
        == "BROADCAST_COMPLIANCE_FORM_PUBLISH"
    )


def test_is_feature_flag_enabled(source: DataHubFormsNotificationsSource) -> None:
    """Test checking if the feature flag is enabled"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Test when feature flag is enabled
    mock_graph.execute_graphql.return_value = {
        "appConfig": {"featureFlags": {"formsNotificationsEnabled": True}}
    }
    assert source.is_feature_flag_enabled() is True

    # Test when feature flag is disabled
    mock_graph.execute_graphql.return_value = {
        "appConfig": {"featureFlags": {"formsNotificationsEnabled": False}}
    }
    assert source.is_feature_flag_enabled() is False

    # Test when feature flag is not present
    mock_graph.execute_graphql.return_value = {"appConfig": {"featureFlags": {}}}
    assert source.is_feature_flag_enabled() is False

    # Test when appConfig is not present
    mock_graph.execute_graphql.return_value = {}
    assert source.is_feature_flag_enabled() is False


def test_get_workunits_feature_flag_disabled(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test getting workunits when feature flag is disabled"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock feature flag check to return disabled
    mock_graph.execute_graphql.return_value = {
        "appConfig": {"featureFlags": {"formsNotificationsEnabled": False}}
    }

    workunits = list(source.get_workunits())
    assert len(workunits) == 0  # Should return empty list when feature flag is disabled

    # Verify feature flag was checked
    mock_graph.execute_graphql.assert_called_once_with(
        GRAPHQL_GET_FEATURE_FLAG, variables={}
    )

    # Verify no other methods were called
    mock_graph.get_entities.assert_not_called()


def test_get_workunits_feature_flag_enabled(
    source: DataHubFormsNotificationsSource,
) -> None:
    """Test getting workunits when feature flag is enabled"""
    assert source.graph is not None  # For mypy
    mock_graph = cast(MagicMock, source.graph)

    # Mock form info
    mock_form_info = FormInfoClass(
        name="Test Form",
        type=FormTypeClass.COMPLETION,
        status=FormStatusClass(state=FormStateClass.PUBLISHED),
        actors=FormActorAssignmentClass(users=["urn:li:corpuser:test-user"]),
    )

    # Mock feature flag check to return enabled
    mock_graph.execute_graphql.side_effect = [
        # First call for feature flag check
        {"appConfig": {"featureFlags": {"formsNotificationsEnabled": True}}},
        # Second call for is_form_complete
        {"searchAcrossEntities": {"total": 0}},
    ]

    mock_graph.get_entities.return_value = {
        "urn:li:form:test-form-1": {"formInfo": (mock_form_info, None)}
    }

    workunits = list(source.get_workunits())
    assert len(workunits) == 0  # This source doesn't produce work units

    # Verify feature flag was checked
    mock_graph.execute_graphql.assert_any_call(GRAPHQL_GET_FEATURE_FLAG, variables={})

    # Verify form info was retrieved
    mock_graph.get_entities.assert_called_once_with(
        "form", ["urn:li:form:test-form-1"], ["formInfo"]
    )



import logging
import math
import os
import re
import tempfile
import time
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from functools import partial
from itertools import chain
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import numpy
import polars
import pyarrow as pa
import pyarrow.parquet as pq
from elasticsearch.client import Elasticsearch
from opensearchpy import OpenSearch
from polars.datatypes import DataTypeClass
from pydantic import Field
from scipy.stats import expon

from acryl_datahub_cloud.datahub_usage_reporting.query_builder import QueryBuilder
from acryl_datahub_cloud.datahub_usage_reporting.usage_feature_patch_builder import (
    UsageFeaturePatchBuilder,
)
from acryl_datahub_cloud.elasticsearch.config import ElasticSearchClientConfig
from acryl_datahub_cloud.metadata.schema_classes import (
    QueryUsageFeaturesClass,
    UsageFeaturesClass,
)
from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceReport
from datahub.ingestion.api.source_helpers import auto_workunit_reporter
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DatahubClientConfig
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionReport,
    StatefulIngestionSourceBase,
)
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.utilities.perf_timer import PerfTimer

logger = logging.getLogger(__name__)

platform_regexp = re.compile(r"urn:li:dataset:\(urn:li:dataPlatform:(.+?),.*")
dashboard_chart_platform_regexp = re.compile(r"urn:li:(?:dashboard|chart):\((.+?),.*")
dbt_platform_regexp = re.compile(r"urn:li:dataset:\(urn:li:dataPlatform:dbt,.*\)")


class S3ClientConfig(ConfigModel):
    bucket: str = os.getenv("DATA_BUCKET", "")
    path: str = os.getenv("RDS_DATA_PATH", "rds_backup/metadata_aspect_v2")


class FreshnessFactor(ConfigModel):
    age_in_days: List[int]
    value: float


class RegexpFactor(ConfigModel):
    regexp: str
    value: float


class UsagePercentileFactor(ConfigModel):
    percentile: List[int]
    value: float


@dataclass
class SearchRankingMultipliers:
    usageSearchScoreMultiplier: Optional[float] = 1.0
    usageFreshnessScoreMultiplier: Optional[float] = 1.0
    customDatahubScoreMultiplier: Optional[float] = 1.0
    combinedSearchRankingMultiplier: Optional[float] = 1.0


class RankingPolicy(ConfigModel):
    freshness_factors: List[FreshnessFactor] = []
    usage_percentile_factors: List[UsagePercentileFactor] = []
    regexp_based_factors: List[RegexpFactor] = []


class DataHubUsageFeatureReportingSourceConfig(
    ConfigModel, StatefulIngestionConfigBase
):
    lookback_days: int = Field(
        30, description="Number of days to look back for usage data."
    )

    server: Optional[DatahubClientConfig] = Field(
        None, description="Optional configuration for the DataHub server connection."
    )
    search_index: ElasticSearchClientConfig = Field(
        default_factory=ElasticSearchClientConfig,
        description="Configuration for the Elasticsearch or OpenSearch index.",
    )
    query_timeout: int = Field(
        30, description="Timeout in seconds for the search queries."
    )
    extract_batch_size: int = Field(
        5000,
        description="The number of documents to retrieve in each batch from ElasticSearch or OpenSearch.",
    )

    extract_delay: Optional[float] = Field(
        0,
        description="The delay in seconds between each batch extraction from ElasticSearch or OpenSearch.",
    )

    use_exp_cdf: bool = Field(
        True,
        description="Flag to determine whether to use the exponential cumulative distribution function for calculating percentiles.",
    )
    ranking_policy: RankingPolicy = Field(
        default_factory=RankingPolicy,
        description="Configuration for the ranking policy.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        None,
        description="Optional configuration for stateful ingestion, including stale metadata removal.",
    )
    dataset_usage_enabled: bool = Field(
        True,
        description="Flag to enable or disable dataset usage statistics collection.",
    )
    dashboard_usage_enabled: bool = Field(
        True,
        description="Flag to enable or disable dashboard usage statistics collection.",
    )
    chart_usage_enabled: bool = Field(
        True, description="Flag to enable or disable chart usage statistics collection."
    )

    query_usage_enabled: bool = Field(
        default=False,
        description="Flag to enable or disable query usage statistics collection.",
    )

    sibling_usage_enabled: bool = Field(
        True,
        description="Flag to enable or disable the setting dataset usage statistics for sibling entities (only DBT siblings are set).",
    )

    use_server_side_aggregation: bool = Field(
        False,
        description="Flag to enable server side aggregation for write usage statistics.",
    )

    set_upstream_table_max_modification_time_for_views: bool = Field(
        True,
        description="Flag to enable setting the max modification time for views based on their upstream tables' modification time.'",
    )

    streaming_mode: bool = Field(
        True,
        description="Flag to enable polars streaming mode.'",
    )

    # Running the whole pipeline in streaming mode was very unstable in the past.
    # It seems like with the latest version of Polars it is much more stable.
    # This option is only needed here until we are sure that the streaming mode is stable.
    # then we can remove it and control it with the streaming_mode option.
    experimental_full_streaming: bool = Field(
        True,
        description="Flag to enable full streaming mode.'",
    )

    disable_write_usage: bool = Field(
        True,
        description="Flag to disable write usage statistics collection.'",
    )

    generate_patch: bool = Field(
        True,
        description="Flag to generate MCP patch for usage features.'",
    )


def exp_cdf(series: polars.Series) -> polars.Series:
    with PerfTimer() as timer:
        if series.is_empty():
            return polars.Series([])

        numpy_array = series.to_numpy()
        fit_array = numpy_array[~numpy.isnan(numpy_array)]
        if fit_array.size == 0:
            return polars.Series([0] * len(numpy_array))

        loc, scale = expon.fit(fit_array, floc=0)
        # percentiles = [int(round(expon.cdf(count, loc, scale) * 100)) for count in numpy_array]
        percentiles = []
        for count in numpy_array:
            if math.isnan(count):
                percentiles.append(0)
            else:
                try:
                    exp_cdf_value = round(expon.cdf(count, loc, scale) * 100)
                except Exception as e:
                    logger.warning(
                        f"Expcdf calculation failed on array: {numpy_array}, count: {count}, loc: {loc}, scale: {scale}, Error: {e}"
                    )
                    return polars.Series([0] * len(numpy_array))
                if math.isnan(exp_cdf_value):
                    percentiles.append(0)
                else:
                    percentiles.append(int(exp_cdf_value))

        logger.debug(f"Percentiles: {percentiles}")
        time_taken = timer.elapsed_seconds()
        logger.debug(f"Exp CDF processing took {time_taken:.3f} seconds")
    return polars.Series(percentiles)


@dataclass
class DatahubUsageFeatureReport(IngestionStageReport, StatefulIngestionReport):
    dataset_platforms_count: Dict[str, int] = field(
        default_factory=lambda: defaultdict(lambda: 0)
    )
    dashboard_platforms_count: Dict[str, int] = field(
        default_factory=lambda: defaultdict(lambda: 0)
    )
    sibling_usage_count: int = 0

    report_es_extraction_time: Dict[str, PerfTimer] = field(
        default_factory=lambda: defaultdict(lambda: PerfTimer())
    )

    dataset_usage_processing_time: PerfTimer = PerfTimer()
    dashboard_usage_processing_time: PerfTimer = PerfTimer()
    chart_usage_processing_time: PerfTimer = PerfTimer()
    query_usage_processing_time: PerfTimer = PerfTimer()
    query_platforms_count: Dict[str, int] = field(
        default_factory=lambda: defaultdict(lambda: 0)
    )


@platform_name(id="datahub", platform_name="DataHub")
@config_class(DataHubUsageFeatureReportingSourceConfig)
@support_status(SupportStatus.INCUBATING)
class DataHubUsageFeatureReportingSource(StatefulIngestionSourceBase):
    platform = "datahub"
    temp_files_to_clean: List[str] = []
    temp_dir: Optional[TemporaryDirectory] = None

    def __init__(
        self, ctx: PipelineContext, config: DataHubUsageFeatureReportingSourceConfig
    ):
        super().__init__(config, ctx)
        # super().__init__(ctx)
        self.config: DataHubUsageFeatureReportingSourceConfig = config
        self.report: DatahubUsageFeatureReport = DatahubUsageFeatureReport()
        self.ctx = ctx

        # We compile regexpes in advance for faster matching
        self.compiled_regexp_factor: List[Tuple[re.Pattern[str], float]] = []
        num = 0
        for rfactor in self.config.ranking_policy.regexp_based_factors:
            self.compiled_regexp_factor.append(
                (re.compile(rfactor.regexp), rfactor.value)
            )
            num += 1

        if num > 0:
            logger.info(f"Compiled {num} regexp factors")

        if self.config.streaming_mode:
            self.temp_dir = tempfile.TemporaryDirectory(prefix="datahub-usage-")
            logger.info(f"Using temp dir: {self.temp_dir.name}")

    def soft_deleted_batch(self, results: Iterable) -> Iterable[Dict]:
        with PerfTimer() as timer:
            for doc in results:
                try:
                    if "urn" not in doc["_source"]:
                        logger.warning(f"Urn not found in ES doc {doc}. Skipping...")
                        continue

                    yield {
                        "entity_urn": doc["_source"]["urn"],
                        "last_modified_at": (
                            doc["_source"]["lastModifiedAt"]
                            if "lastModifiedAt" in doc["_source"]
                            and doc["_source"]["lastModifiedAt"]
                            else (
                                doc["_source"]["lastModifiedAt"]
                                if "lastModifiedAt" in doc["_source"]
                                and doc["_source"]["lastModifiedAt"]
                                else None
                            )
                        ),
                        "removed": (
                            doc["_source"]["removed"]
                            if "removed" in doc["_source"] and doc["_source"]["removed"]
                            else False
                        ),
                        "siblings": (
                            doc["_source"]["siblings"]
                            if "siblings" in doc["_source"]
                            and doc["_source"]["siblings"]
                            else []
                        ),
                        "combinedSearchRankingMultiplier": (
                            doc["_source"]["combinedSearchRankingMultiplier"]
                            if "combinedSearchRankingMultiplier" in doc["_source"]
                            and doc["_source"]["combinedSearchRankingMultiplier"]
                            else None
                        ),
                        "isView": (
                            "View" in doc["_source"]["typeNames"]
                            if "typeNames" in doc["_source"]
                            and doc["_source"]["typeNames"]
                            else False
                        ),
                    }
                except KeyError as e:
                    logger.warning(
                        f"Unable to process row {doc} from ES. It failed with {e}"
                    )
                    continue
            time_taken = timer.elapsed_seconds()
            logger.info(f"Entities processing took {time_taken:.3f} seconds")

    def write_stat_batch(self, results: Iterable) -> Iterable[Dict]:
        with PerfTimer() as timer:
            for doc in results:
                match = re.match(
                    platform_regexp, doc["key"]["dataset_operationaspect_v1"]
                )
                if match:
                    platform = match.group(1)
                else:
                    logger.warning("Platform not found in urn. Skipping...")
                    continue

                yield {
                    "urn": doc["key"]["dataset_operationaspect_v1"],
                    "platform": platform,
                    "write_count": doc["doc_count"],
                }
            time_taken = timer.elapsed_seconds()
            logger.info(
                f"Write Operation aspect processing took {time_taken:.3f} seconds"
            )

    def write_stat_raw_batch(self, results: Iterable) -> Iterable[Dict]:
        with PerfTimer() as timer:
            for doc in results:
                match = re.match(platform_regexp, doc["_source"]["urn"])
                if match:
                    platform = match.group(1)
                else:
                    logger.warning("Platform not found in urn. Skipping...")
                    continue

                yield {
                    "urn": doc["_source"]["urn"],
                    "platform": platform,
                }
            time_taken = timer.elapsed_seconds()
            logger.info(
                f"Write Operation aspect processing took {time_taken:.3f} seconds"
            )

    def queries_entities_batch(self, results: Iterable) -> Iterable[Dict]:
        with PerfTimer() as timer:
            for doc in results:
                if "platform" not in doc["_source"] or not doc["_source"]["platform"]:
                    logger.warning(
                        f"Platform not found in query {doc['_source']['urn']}. Skipping..."
                    )
                    continue

                self.report.query_platforms_count[doc["_source"]["platform"]] = (
                    self.report.query_platforms_count[doc["_source"]["platform"]] + 1
                )

                yield {
                    "entity_urn": doc["_source"]["urn"],
                    "last_modified_at": (
                        doc["_source"]["lastModifiedAt"]
                        if "lastModifiedAt" in doc["_source"]
                        else (doc["_source"].get("lastModifiedAt", None))
                    ),
                    "platform": doc["_source"]["platform"],
                    "removed": (doc["_source"].get("removed", False)),
                }

            time_taken = timer.elapsed_seconds()
            logger.info(f"Query entities processing took {time_taken:.3f} seconds")

    def process_dashboard_usage(self, results: Iterable) -> Iterable[Dict]:
        for doc in results:
            match = re.match(dashboard_chart_platform_regexp, doc["_source"]["urn"])
            if match:
                platform = match.group(1)
                self.report.dashboard_platforms_count[platform] += 1
            else:
                logger.warning("Platform not found in urn. Skipping...")
                continue

            yield {
                "timestampMillis": doc["_source"].get("timestampMillis"),
                "lastObserved": doc["_source"]
                .get("systemMetadata", {})
                .get("lastObserved"),
                "urn": doc["_source"].get("urn"),
                "eventGranularity": doc["_source"].get("eventGranularity"),
                "viewsCount": doc["_source"].get("viewsCount", 0),
                "uniqueUserCount": doc["_source"].get("uniqueUserCount"),
                "userCounts": doc["_source"].get("event", {}).get("userCounts", []),
                "platform": platform,
            }

    def process_query_usage(self, results: Iterable) -> Iterable[Dict]:
        for doc in results:
            yield {
                "timestampMillis": doc["_source"].get("timestampMillis"),
                "lastObserved": doc["_source"]
                .get("systemMetadata", {})
                .get("lastObserved"),
                "urn": doc["_source"].get("urn"),
                "eventGranularity": doc["_source"].get("eventGranularity"),
                "queryCount": doc["_source"].get("queryCount", 0),
                "uniqueUserCount": doc["_source"].get("uniqueUserCount"),
                "userCounts": doc["_source"].get("event", {}).get("userCounts", []),
            }

    def upstream_lineage_batch(self, results: Iterable) -> Iterable[Dict]:
        for doc in results:
            if (
                not doc["_source"]["source"]["urn"]
                or not doc["_source"]["destination"]["urn"]
            ):
                logger.warning("Source urn not found in upstream lineage. Skipping...")
                continue

            source_platform_match = re.match(
                platform_regexp, doc["_source"]["source"]["urn"]
            )
            if source_platform_match:
                source_platform = source_platform_match.group(1)
            else:
                logger.warning("Source Platform not found in urn. Skipping...")
                continue

            destination_platform_match = re.match(
                platform_regexp, doc["_source"]["destination"]["urn"]
            )
            if destination_platform_match:
                destination_platform = destination_platform_match.group(1)
            else:
                logger.warning("Destination Platform not found in urn. Skipping...")
                continue

            # In some case like Tableau there is dataset which marked as view and points to a dataset on another platform
            # We drop these now
            if source_platform != destination_platform:
                continue

            yield {
                "source_urn": doc["_source"]["source"]["urn"],
                "destination_urn": doc["_source"]["destination"]["urn"],
            }

    def process_batch(self, results: Iterable) -> Iterable[Dict]:
        with PerfTimer() as timer:
            for doc in results:
                if "urn" not in doc["_source"]:
                    logger.warning(f"Urn not found in ES doc {doc}. Skipping...")
                    continue
                match = re.match(platform_regexp, doc["_source"]["urn"])
                if match:
                    platform = match.group(1)
                    self.report.dataset_platforms_count[platform] += 1
                else:
                    logger.warning(
                        f"Platform not found in urn  {doc['_source']['urn']} in doc {doc}. Skipping..."
                    )
                    continue

                try:
                    yield {
                        "timestampMillis": doc["_source"]["timestampMillis"],
                        "urn": doc["_source"]["urn"],
                        "eventGranularity": doc["_source"].get("eventGranularity"),
                        "totalSqlQueries": doc["_source"].get("totalSqlQueries", 0),
                        "uniqueUserCount": doc["_source"].get("uniqueUserCount", 0),
                        "userCounts": (doc["_source"]["event"].get("userCounts", None)),
                        "platform": platform,
                    }
                except KeyError as e:
                    logger.warning(
                        f"Unable to process row {doc} from ES. The error was: {e}"
                    )
                    continue

            time_taken = timer.elapsed_seconds()
            logger.info(f"DatasetUsage processing took {time_taken:.3f} seconds")

    def search_score(  # noqa: C901
        self, urn: str, last_update_time: int, usage_percentile: int
    ) -> SearchRankingMultipliers:
        usage_search_score_multiplier = 1.0
        freshness_factor = 1.0
        regexp_factor = 1.0

        current_time = datetime.now().timestamp() * 1000
        age_in_millis = current_time - last_update_time
        age_in_days = age_in_millis / (1000 * 60 * 60 * 24)

        bucket = 0
        for factor in self.config.ranking_policy.freshness_factors:
            if len(factor.age_in_days) == 2:
                if bucket == 0:
                    if factor.age_in_days[0] <= age_in_days <= factor.age_in_days[1]:
                        freshness_factor = factor.value
                        break
                else:
                    if factor.age_in_days[0] < age_in_days <= factor.age_in_days[1]:
                        freshness_factor = factor.value
                        break
            elif age_in_days > factor.age_in_days[0]:
                freshness_factor = factor.value

        bucket = 0
        for pfactor in self.config.ranking_policy.usage_percentile_factors:
            bucket += 1
            if len(pfactor.percentile) == 2:
                # The first bucket min should be inclusive
                if bucket == 1:
                    if (
                        pfactor.percentile[0]
                        <= usage_percentile
                        <= pfactor.percentile[1]
                    ):
                        usage_search_score_multiplier = pfactor.value
                        break
                else:
                    if (
                        pfactor.percentile[0]
                        < usage_percentile
                        <= pfactor.percentile[1]
                    ):
                        usage_search_score_multiplier = pfactor.value
                        break
            elif usage_percentile > pfactor.percentile[0]:
                usage_search_score_multiplier = pfactor.value

        for rfactor in self.compiled_regexp_factor:
            if rfactor[0].match(urn):
                regexp_factor = rfactor[1]

        return SearchRankingMultipliers(
            usageSearchScoreMultiplier=usage_search_score_multiplier,
            usageFreshnessScoreMultiplier=freshness_factor,
            customDatahubScoreMultiplier=regexp_factor,
            # We make sure the combinedSearchRankingMultiplier is never less than 1
            combinedSearchRankingMultiplier=max(
                1, (usage_search_score_multiplier * freshness_factor * regexp_factor)
            ),
        )

    def load_data_from_es(
        self,
        index: str,
        query: Dict,
        process_function: Callable,
        aggregation_key: Optional[str] = None,
    ) -> Iterable[Dict]:
        with self.report.report_es_extraction_time[index]:
            query_copy = query.copy()
            endpoint = ""
            if self.config.search_index:
                if self.config.search_index.host and not self.config.search_index.port:
                    endpoint = f"{self.config.search_index.host}"
                elif self.config.search_index.host and self.config.search_index.port:
                    endpoint = f"{self.config.search_index.host}:{self.config.search_index.port}"

                index_prefix = (
                    self.config.search_index.index_prefix
                    if self.config.search_index
                    else ""
                )

                index = f"{index_prefix}{index}" if index_prefix else index
                user = self.config.search_index.username
                password = self.config.search_index.password
                batch_size = self.config.extract_batch_size
                delay = self.config.extract_delay
                server: Union[Elasticsearch, OpenSearch]

                if self.config.search_index.opensearch_dialect:
                    server = OpenSearch(
                        [endpoint],
                        http_auth=(user, password),
                        use_ssl=(
                            bool(
                                self.config.search_index
                                and self.config.search_index.use_ssl
                            )
                        ),
                    )

                    # response = server.create_pit(index, keep_alive="10m")

                    # TODO: Save PIT, we can resume processing based on <pit, search_after> tuple
                    # pit = response.get("pit_id")
                    # query_copy.update({"pit": {"id": pit, "keep_alive": "10m"}})
                else:
                    server = Elasticsearch(
                        [endpoint],
                        http_auth=(user, password),
                        use_ssl=(
                            bool(
                                self.config.search_index
                                and self.config.search_index.use_ssl
                            )
                        ),
                    )

                yield from self.load_es_data(
                    query_copy,
                    server,
                    index,
                    process_function,
                    batch_size=batch_size,
                    delay=delay,
                    aggregation_key=aggregation_key,
                )

    def gen_rank_and_percentile(
        self,
        lf: polars.LazyFrame,
        count_field: str,
        urn_field: str = "urn",
        platform_field: str = "platform",
        prefix: Optional[str] = None,
        use_exp_cdf: Optional[bool] = None,
    ) -> polars.LazyFrame:
        logger.debug(f"Generating rank and percentile for {count_field} field")
        lf = lf.with_columns(
            polars.col(count_field)
            .rank(descending=True, method="max")
            .over(platform_field)
            .alias(f"{prefix}rank")
        )

        use_exp_cdf = self.config.use_exp_cdf if use_exp_cdf is None else use_exp_cdf
        if use_exp_cdf:
            lf = lf.with_columns(
                polars.col(count_field)
                .map_batches(exp_cdf, return_dtype=polars.Int64)
                .over(platform_field)
                .alias(f"{prefix}rank_percentile")
            )
        else:
            lf = lf.with_columns(
                polars.when(
                    polars.col(count_field)
                    > 0  # This is slightly modified percentile rank calculation as we zero out zero usage
                )
                .then(
                    (
                        1
                        - (
                            (polars.col(f"{prefix}rank") - 1)
                            / (
                                polars.max_horizontal(
                                    polars.col(urn_field).count(), polars.lit(2)
                                )
                                - 1
                            )  # If we only have 1 item then we have to set the percentile to 100
                        )
                    )
                    * 100
                )
                .otherwise(0)
                .over(platform_field)
                .alias(f"{prefix}rank_percentile")
            )

        return lf

    @staticmethod
    def polars_to_arrow_schema(
        polars_schema: Dict[str, Union[DataTypeClass, polars.DataType]],
    ) -> pa.Schema:
        def convert_dtype(
            polars_dtype: Union[DataTypeClass, polars.DataType],
        ) -> pa.DataType:
            type_mapping: Dict[Union[DataTypeClass, polars.DataType], pa.DataType] = {
                polars.Boolean(): pa.bool_(),
                polars.Int8(): pa.int8(),
                polars.Int16(): pa.int16(),
                polars.Int32(): pa.int32(),
                polars.Int64(): pa.int64(),
                polars.UInt8(): pa.uint8(),
                polars.UInt16(): pa.uint16(),
                polars.UInt32(): pa.uint32(),
                polars.UInt64(): pa.uint64(),
                polars.Float32(): pa.float32(),
                polars.Float64(): pa.float64(),
                polars.Utf8(): pa.string(),
                polars.Utf8(): pa.utf8(),
                polars.String(): pa.string(),
                polars.Date(): pa.date32(),
                polars.Datetime(): pa.timestamp("ns"),
                polars.Time(): pa.time64("ns"),
                polars.Duration(): pa.duration("ns"),
            }

            if polars_dtype in [type(key) for key in type_mapping]:
                return type_mapping[polars_dtype]
            elif polars_dtype == polars.Categorical:
                return pa.dictionary(index_type=pa.int32(), value_type=pa.string())
            elif isinstance(polars_dtype, polars.Struct):
                return pa.struct(
                    {
                        field.name: convert_dtype(field.dtype)
                        for field in polars_dtype.fields
                    }
                )
            elif isinstance(polars_dtype, polars.List):
                return pa.list_(convert_dtype(polars_dtype.inner))
            else:
                raise ValueError(f"Unsupported Polars dtype: {polars_dtype}")

        fields = [(name, convert_dtype(dtype)) for name, dtype in polars_schema.items()]
        return pa.schema(fields)

    def batch_write_parquet(
        self,
        data_iterator: Iterable[Dict[Any, Any]],
        pl_schema: Dict,
        output_path: str,
        batch_size: int = 50000,
        append: bool = False,
        parquet_writer: Optional[pq.ParquetWriter] = None,
    ) -> None:
        """
        Write data in batches to a file with support for appending to existing files.

        Args:
            data_iterator: Iterator of dictionaries containing the data
            pa_schema: PyArrow schema for the data
            output_path: Path for the output file
            format_type: One of "ipc", "feather", "csv", "parquet", "pl_parquet"
            batch_size: Number of rows per batch
            append: If True, append to existing file. If False, create new file.
            parquet_writer: Parquet doesn't let to append to existing file, so we need to pass the writer object
        Returns:
            LazyFrame pointing to the written data
        """
        arrow_schema = self.polars_to_arrow_schema(pl_schema)

        total_rows = 0
        total_batches = 0

        try:
            if parquet_writer:
                writer = parquet_writer
            else:
                writer = pq.ParquetWriter(output_path, arrow_schema)

            try:
                for batch in self._get_batches(data_iterator, batch_size):
                    table = pa.Table.from_pylist(batch, schema=arrow_schema)
                    writer.write_table(table)
                    total_rows += len(batch)
                    total_batches += 1
                    logger.debug(f"Wrote batch {total_batches} ({len(batch)} rows)")
            finally:
                if not parquet_writer:
                    writer.close()
        except Exception as e:
            logger.exception(f"Error during batch writing: {str(e)}", exc_info=True)
            raise

    def _get_batches(
        self, iterator: Iterable[Dict], batch_size: int
    ) -> Iterator[List[Dict]]:
        """Helper generator to create batches from an iterator."""
        current_batch = []
        for item in iterator:
            current_batch.append(item)
            if len(current_batch) >= batch_size:
                yield current_batch
                current_batch = []

        if current_batch:
            yield current_batch

    def load_write_usage(
        self, soft_deleted_entities_df: polars.LazyFrame
    ) -> polars.LazyFrame:
        wdf = self.load_data_from_es_to_lf(
            index="dataset_operationaspect_v1",
            query=QueryBuilder.get_dataset_write_usage_raw_query(
                self.config.lookback_days
            ),
            process_function=self.write_stat_raw_batch,
            schema={"urn": polars.Categorical, "platform": polars.Categorical},
        )
        wdf = wdf.cast({polars.String: polars.Categorical})

        wdf = wdf.group_by(polars.col("urn"), polars.col("platform")).agg(
            polars.col("urn").count().alias("write_count"),
        )

        wdf = (
            wdf.join(
                soft_deleted_entities_df,
                left_on="urn",
                right_on="entity_urn",
                how="inner",
            )
            .filter(polars.col("removed") == False)  # noqa: E712
            .drop(["removed"])
        )

        return wdf

    def load_write_usage_server_side_aggregation(
        self, soft_deleted_entities_df: polars.LazyFrame
    ) -> polars.LazyFrame:
        wdf = polars.LazyFrame(
            self.load_data_from_es(
                "dataset_operationaspect_v1",
                QueryBuilder.get_dataset_write_usage_composite_query(
                    self.config.lookback_days
                ),
                self.write_stat_batch,
                aggregation_key="urn_count",
            ),
            schema={
                "urn": polars.Categorical,
                "platform": polars.Categorical,
                "write_count": polars.Int64,
            },
            strict=True,
        )

        wdf = (
            wdf.join(
                soft_deleted_entities_df,
                left_on="urn",
                right_on="entity_urn",
                how="inner",
            )
            .filter(polars.col("removed") == False)  # noqa: E712
            .drop(["removed"])
        )

        return wdf

    def set_table_modification_time_for_views(
        self, datasets_df: polars.LazyFrame
    ) -> polars.LazyFrame:
        schema = {
            "source_urn": polars.Categorical,
            "destination_urn": polars.Categorical,
        }

        upstreams_lf = self.load_data_from_es_to_lf(
            schema=schema,
            index="graph_service_v1",
            query=QueryBuilder.get_upstreams_query(),
            process_function=self.upstream_lineage_batch,
        )

        wdf = (
            (
                upstreams_lf.join(
                    datasets_df.filter(polars.col("isView") == True),  # noqa: E712
                    left_on="destination_urn",
                    right_on="entity_urn",
                    how="inner",
                )
            )
            .join(
                datasets_df.filter(polars.col("isView") == False),  # noqa: E712
                left_on="source_urn",
                right_on="entity_urn",
            )
            .group_by(
                "destination_urn",
            )
            .agg(
                polars.col("last_modified_at_right")
                .max()
                .alias("inherited_last_modified_at"),
                polars.col("last_modified_at").first().alias("last_modified_at"),
            )
        )

        dataset_df = (
            datasets_df.join(
                wdf, left_on="entity_urn", right_on="destination_urn", how="left"
            )
            .with_columns(
                polars.coalesce("inherited_last_modified_at", "last_modified_at").alias(
                    "last_modified_at"
                )
            )
            .drop(["inherited_last_modified_at", "last_modified_at_right"])
        )

        return dataset_df

    def generate_dataset_usage_mcps(self) -> Iterable[MetadataWorkUnit]:
        with polars.StringCache():
            dataset_usage_df = self.generate_dataset_usage()
            logger.info("Generate Dataset Usage")
            yield from self.generate_mcp_from_lazyframe(dataset_usage_df)
            logger.info("End Generate Dataset Usage")

    def generate_dashboard_usage_mcps(self) -> Iterable[MetadataWorkUnit]:
        with polars.StringCache():
            logger.info("Generate Dashboard Usage")
            dashboard_usage_df = self.generate_dashboard_usage()
            yield from self.generate_mcp_from_lazyframe(dashboard_usage_df)

    def generate_chart_usage_mcps(self) -> Iterable[MetadataWorkUnit]:
        with polars.StringCache():
            logger.info("Generate Chart Usage")
            chart_usage_df = self.generate_chart_usage()
            yield from self.generate_mcp_from_lazyframe(chart_usage_df)

    def generate_query_usage_mcps(self) -> Iterable[MetadataWorkUnit]:
        with polars.StringCache():
            logger.info("Generate Query Usage")
            query_usage_df = self.generate_query_usage()
            yield from self.generate_query_usage_mcp_from_lazyframe(query_usage_df)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        """A list of functions that transforms the workunits produced by this source.
        Run in order, first in list is applied first. Be careful with order when overriding.
        """

        return [
            partial(auto_workunit_reporter, self.get_report()),
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        if self.config.dataset_usage_enabled:
            with self.report.dataset_usage_processing_time as timer:
                self.report.new_stage("generate dataset usage")
                yield from self.generate_dataset_usage_mcps()
                time_taken = timer.elapsed_seconds()
                logger.info(f"Dataset Usage generation took {time_taken:.3f} seconds")

        if self.config.dashboard_usage_enabled:
            with self.report.dashboard_usage_processing_time as timer:
                self.report.new_stage("generate dashboard usage")
                yield from self.generate_dashboard_usage_mcps()

                time_taken = timer.elapsed_seconds()
                logger.info(f"Dashboard Usage generation took {time_taken:.3f}")

        if self.config.chart_usage_enabled:
            with self.report.chart_usage_processing_time as timer:
                self.report.new_stage("generate chart usage")

                yield from self.generate_chart_usage_mcps()

                time_taken = timer.elapsed_seconds()
                logger.info(f"Chart Usage generation took {time_taken:.3f}")

        if self.config.query_usage_enabled:
            with self.report.query_usage_processing_time as timer:
                self.report.new_stage("generate query usage")

                yield from self.generate_query_usage_mcps()

                time_taken = timer.elapsed_seconds()
                logger.info(f"Query Usage generation took {time_taken:.3f}")

    def generate_mcp_from_lazyframe(
        self, lazy_frame: polars.LazyFrame
    ) -> Iterable[MetadataWorkUnit]:
        for row in lazy_frame.collect(
            engine="streaming" if self.config.experimental_full_streaming else "auto"
        ).to_struct():
            if "siblings" in row and row["siblings"]:
                logger.info(f"Siblings found for urn: {row['urn']} -> row['siblings']")

            search_ranking_multipliers: SearchRankingMultipliers = (
                SearchRankingMultipliers()
            )

            if "queries_rank_percentile" in row:
                # If usage data is missing we set the search ranking multipliers to 1
                search_ranking_multipliers = (
                    self.search_score(
                        urn=row["urn"],
                        last_update_time=row.get("last_modified_at", 0) or 0,
                        usage_percentile=row.get("queries_rank_percentile", 0) or 0,
                    )
                    if row.get("queries_rank_percentile", 0)
                    else SearchRankingMultipliers()
                )
            elif "viewsCount30Days_rank_percentile" in row:
                # If usage data is missing we set the search ranking multipliers to 1
                search_ranking_multipliers = (
                    self.search_score(
                        urn=row["urn"],
                        last_update_time=row.get("last_modified_at", 0) or 0,
                        usage_percentile=row.get("viewsCount30Days_rank_percentile", 0)
                        or 0,
                    )
                    if row.get("viewsCount30Days_rank_percentile", 0)
                    else SearchRankingMultipliers()
                )
                logger.debug(f"Urn: {row['urn']} Score: {search_ranking_multipliers}")

            usage_feature = UsageFeaturesClass(
                queryCountLast30Days=int(row.get("totalSqlQueries", 0) or 0),
                usageCountLast30Days=int(row.get("totalSqlQueries", 0) or 0),
                queryCountRankLast30Days=int(row.get("queries_rank"))
                if row.get("queries_rank")
                else None,
                queryCountPercentileLast30Days=row.get("queries_rank_percentile", 0)
                or 0,
                # queryCountPercentileLast30Days=int(
                #   row["queries_rank_percentile"]) if "queries_rank_percentile" in row and row[
                #   "queries_rank_percentile"] else 0,
                topUsersLast30Days=(
                    list(chain.from_iterable(row.get("top_users")))
                    if row.get("top_users")
                    else None
                ),
                uniqueUserCountLast30Days=int(row.get("distinct_user", 0) or 0),
                uniqueUserRankLast30Days=int(row.get("distinct_user_rank"))
                if row.get("distinct_user_rank")
                else None,
                uniqueUserPercentileLast30Days=int(
                    row.get("distinct_user_rank_percentile", 0) or 0
                ),
                writeCountLast30Days=int(row.get("write_rank_percentile", 0) or 0)
                if not self.config.disable_write_usage
                else None,
                writeCountPercentileLast30Days=int(
                    row.get("write_rank_percentile", 0) or 0
                )
                if not self.config.disable_write_usage
                else None,
                writeCountRankLast30Days=int(row.get("write_rank") or 0)
                if not self.config.disable_write_usage
                else None,
                viewCountTotal=int(row.get("viewsTotal", 0) or 0),
                viewCountLast30Days=int(row.get("viewsCount30Days", 0) or 0),
                viewCountPercentileLast30Days=int(
                    row.get("viewsCount30Days_rank_percentile", 0) or 0
                ),
                usageSearchScoreMultiplier=search_ranking_multipliers.usageSearchScoreMultiplier,
                usageFreshnessScoreMultiplier=search_ranking_multipliers.usageFreshnessScoreMultiplier,
                customDatahubScoreMultiplier=search_ranking_multipliers.customDatahubScoreMultiplier,
                combinedSearchRankingMultiplier=search_ranking_multipliers.combinedSearchRankingMultiplier,
            )

            yield from self.generate_usage_feature_mcp(row["urn"], usage_feature)

            if row.get("siblings") and self.config.sibling_usage_enabled:
                for sibling in row["siblings"]:
                    if dbt_platform_regexp.match(sibling):
                        yield from self.generate_usage_feature_mcp(
                            sibling, usage_feature
                        )

    def generate_query_usage_mcp_from_lazyframe(
        self, lazy_frame: polars.LazyFrame
    ) -> Iterable[MetadataWorkUnit]:
        for row in lazy_frame.collect(
            engine="streaming" if self.config.experimental_full_streaming else "auto"
        ).iter_rows(named=True):
            query_usage_features = QueryUsageFeaturesClass(
                queryCountLast30Days=int(row.get("totalSqlQueries", 0) or 0),
                queryCountTotal=None,  # This is not implemented
                runsPercentileLast30days=int(
                    row.get("queries_rank_percentile", 0) or 0
                ),
                lastExecutedAt=int(row.get("last_modified_at", 0) or 0),
                topUsersLast30Days=(
                    list(chain.from_iterable(row.get("top_users", [])))
                    if row.get("top_users")
                    else None
                ),
                queryCostLast30Days=None,  # Not implemented yet
            )

            yield from self.generate_query_usage_feature_mcp(
                row["urn"], query_usage_features
            )

    def generate_usage_feature_mcp(
        self, urn: str, usage_feature: UsageFeaturesClass
    ) -> Iterable[MetadataWorkUnit]:
        if self.config.generate_patch:
            usage_feature_patch_builder = UsageFeaturePatchBuilder(urn=urn)
            usage_feature_patch_builder.apply_usage_features(usage_feature)
            for mcp in usage_feature_patch_builder.build():
                yield MetadataWorkUnit(
                    id=MetadataWorkUnit.generate_workunit_id(mcp),
                    mcp_raw=mcp,
                    is_primary_source=False,
                )
        else:
            mcw = MetadataChangeProposalWrapper(entityUrn=urn, aspect=usage_feature)
            yield mcw.as_workunit(is_primary_source=False)

    def generate_query_usage_feature_mcp(
        self, urn: str, query_usage_features: QueryUsageFeaturesClass
    ) -> Iterable[MetadataWorkUnit]:
        mcw = MetadataChangeProposalWrapper(entityUrn=urn, aspect=query_usage_features)
        yield mcw.as_workunit(is_primary_source=False)

    def generate_chart_usage(self) -> polars.LazyFrame:
        entity_index = "chartindex_v2"
        usage_index = "chart_chartusagestatisticsaspect_v1"

        return self.generate_dashboard_chart_usage(entity_index, usage_index)

    def generate_dashboard_usage(self) -> polars.LazyFrame:
        usage_index = "dashboard_dashboardusagestatisticsaspect_v1"
        entity_index = "dashboardindex_v2"

        return self.generate_dashboard_chart_usage(entity_index, usage_index)

    def generate_dashboard_chart_usage(
        self, entity_index: str, usage_index: str
    ) -> polars.LazyFrame:
        entity_schema = {
            "entity_urn": polars.Categorical,
            "removed": polars.Boolean,
            "last_modified_at": polars.Int64,
            "siblings": polars.List(polars.String),
            "combinedSearchRankingMultiplier": polars.Float64,
            "isView": polars.Boolean,
        }

        entities_df = self.load_data_from_es_to_lf(
            schema=entity_schema,
            index=entity_index,
            query=QueryBuilder.get_dataset_entities_query(),
            process_function=self.soft_deleted_batch,
        )

        dashboard_usage_schema = {
            "timestampMillis": polars.Int64,
            "lastObserved": polars.Int64,
            "urn": polars.Categorical,
            "platform": polars.Categorical,
            "eventGranularity": polars.String,
            "viewsCount": polars.Int64,
            "userCounts": polars.List(
                polars.Struct(
                    {
                        "usageCount": polars.Int64,
                        "user": polars.String,
                    }
                )
            ),
        }

        lf = self.load_data_from_es_to_lf(
            schema=dashboard_usage_schema,
            index=usage_index,
            query=QueryBuilder.get_dashboard_usage_query(self.config.lookback_days),
            process_function=self.process_dashboard_usage,
        )

        lf = (
            lf.join(entities_df, left_on="urn", right_on="entity_urn", how="inner")
            .filter(polars.col("removed") == False)  # noqa: E712
            .drop(["removed"])
        )

        lf = lf.with_columns(
            polars.col("lastObserved")
            .rank(descending=True, method="ordinal")
            .over("urn", "timestampMillis")
            .alias("row_num")
        ).filter(polars.col("row_num") == 1)

        # lf = lf.filter(polars.col("urn") == "urn:li:dashboard:(looker,dashboards.8)")
        # "urn:li:dashboard:(looker,dashboards.8)"

        top_users = self.generate_top_users(
            lf.filter(polars.col("eventGranularity").is_not_null()),
            count_field_name="usageCount",
        )

        views_sum_with_top_users = (
            lf.group_by("urn")
            .agg(
                [
                    polars.max("last_modified_at").alias("last_modified_at"),
                    polars.first("siblings").alias("siblings"),
                ]
            )
            .join(top_users, on="urn", how="left")
        )
        # views_sum_with_top_users = views_sum_with_top_users.drop(["userCounts"])

        incremental_views_sum = (
            lf.filter(polars.col("eventGranularity").is_null())
            .group_by("urn")
            .agg(
                polars.col("viewsCount").min().alias("first_viewsCount"),
                polars.col("viewsCount").max().alias("viewsTotal"),
                polars.col("last_modified_at").max().alias("last_modified_at"),
                polars.col("siblings").first().alias("siblings"),
            )
            .with_columns(
                (polars.col("viewsTotal") - polars.col("first_viewsCount")).alias(
                    "viewsCountTotal30Days"
                )
            )
            .drop(["first_viewsCount"])
        )
        views_with_inceremental_sum = views_sum_with_top_users.join(
            incremental_views_sum, on="urn", how="left"
        )
        total_views = views_with_inceremental_sum.with_columns(
            polars.when(
                polars.col("total_user_count")
                .is_null()
                .or_(polars.col("total_user_count") <= 0)
            )
            .then(polars.col("viewsCountTotal30Days"))
            .otherwise(polars.col("total_user_count"))
            .alias("viewsCount30Days")
        )

        total_views_with_rank_and_percentiles = self.gen_rank_and_percentile(
            total_views, "viewsCount30Days", "urn", "platform", "viewsCount30Days_"
        ).drop(["siblings_right"])

        total_views_with_rank_and_percentiles_with_zeroed_stale_usages = (
            self.generate_empty_usage_for_stale_entities(
                entities_df, total_views_with_rank_and_percentiles
            )
        )

        return total_views_with_rank_and_percentiles_with_zeroed_stale_usages

    def generate_empty_usage_for_stale_entities(
        self, entities_lf: polars.LazyFrame, usages_lf: polars.LazyFrame
    ) -> polars.LazyFrame:
        # We need to merge datasets with existing search scores to make sure we can downrank them if there were no usage in the last n days
        # We drop last_modified_at to not use it in merge because we are getting last_modified_at from the usage index
        df_with_search_scores = (
            entities_lf.filter(
                polars.col("combinedSearchRankingMultiplier")
                .is_not_null()
                # We only want to downrank datasets that have a search score multiplier greater than 1. 1 is the minimum score of a dataset
                .and_(polars.col("combinedSearchRankingMultiplier").ne(1))
            )
            .filter(polars.col("removed") == False)  # noqa: E712
            .drop(["removed"])
            .drop(["last_modified_at"])
            # We set this to 0 because we want to downrank datasets that have no usage
            .with_columns(polars.lit(0).alias("combinedSearchRankingMultiplier"))
            .rename({"entity_urn": "urn"})
        )
        common_fields = list(
            set(usages_lf.columns).intersection(set(df_with_search_scores.columns))
        )
        usages_lf = df_with_search_scores.join(
            usages_lf, on="urn", how="full", suffix="_right"
        )
        ## Merge all common fields automatically
        for common_field in common_fields:
            right_col = f"{common_field}_right"
            usages_lf = usages_lf.with_columns(
                [
                    polars.col(common_field)
                    .fill_null(polars.col(right_col))
                    .alias(common_field)
                ]
            ).drop(right_col)
        return usages_lf

    def generate_query_usage(self) -> polars.LazyFrame:
        usage_index = "query_queryusagestatisticsaspect_v1"
        entity_index = "queryindex_v2"
        query_entities_schema = {
            "entity_urn": polars.Categorical,
            "last_modified_at": polars.Int64,
            "platform": polars.Categorical,
            "removed": polars.Boolean,
        }

        query_entities = self.load_data_from_es_to_lf(
            schema=query_entities_schema,
            index=entity_index,
            query=QueryBuilder.get_query_entities_query(self.config.lookback_days),
            process_function=self.queries_entities_batch,
        )

        query_usage_schema = {
            "timestampMillis": polars.Int64,
            "lastObserved": polars.Int64,
            "urn": polars.Categorical,
            "eventGranularity": polars.String,
            "queryCount": polars.Int64,
            "userCounts": polars.List(
                polars.Struct(
                    {
                        "usageCount": polars.Int64,
                        "user": polars.String,
                    }
                )
            ),
        }

        lf = self.load_data_from_es_to_lf(
            schema=query_usage_schema,
            index=usage_index,
            query=QueryBuilder.get_query_usage_query(self.config.lookback_days),
            process_function=self.process_query_usage,
        )

        lf = query_entities.join(
            lf, left_on="entity_urn", right_on="urn", how="left", coalesce=False
        ).filter(
            polars.col("removed") == False  # noqa: E712
        )

        total_queries = lf.group_by("urn", "platform").agg(
            polars.col("queryCount").sum().alias("totalSqlQueries"),
            polars.col("last_modified_at").max().alias("last_modified_at"),
        )

        top_users = self.generate_top_users(lf, "usageCount")

        usage_with_top_users = top_users.join(total_queries, on="urn", how="inner")

        usage_with_top_users_with_ranks = self.gen_rank_and_percentile(
            lf=usage_with_top_users,
            count_field="totalSqlQueries",
            urn_field="urn",
            platform_field="platform",
            prefix="queries_",
            use_exp_cdf=False,
        )

        usage_with_top_users_with_ranks = usage_with_top_users_with_ranks.sort(
            by=["platform", "queries_rank"], descending=[False, False]
        )

        return usage_with_top_users_with_ranks

    def generate_dataset_usage(self) -> polars.LazyFrame:
        datasets_lf = self.get_datasets()
        if self.config.set_upstream_table_max_modification_time_for_views:
            datasets_lf = self.set_table_modification_time_for_views(datasets_lf)

        lf = self.load_dataset_usage()

        # Polaris/pandas join merges the join column into one column and that's why we need to filter based on the removed column
        lf = (
            lf.join(datasets_lf, left_on="urn", right_on="entity_urn", how="left")
            .filter(polars.col("removed") == False)  # noqa: E712
            .drop(["removed"])
        )

        total_queries = lf.group_by("urn", "platform").agg(
            polars.col("totalSqlQueries").sum(),
            polars.col("last_modified_at").max().alias("last_modified_at"),
            polars.col("siblings").first().alias("siblings"),
        )

        total_queries = self.generate_empty_usage_for_stale_entities(
            datasets_lf, total_queries
        )

        top_users = self.generate_top_users(lf)

        usage_with_top_users = total_queries.join(top_users, on="urn", how="left")

        usage_with_top_users_with_ranks = self.gen_rank_and_percentile(
            usage_with_top_users, "totalSqlQueries", "urn", "platform", "queries_"
        )

        usage_with_top_users_with_ranks = usage_with_top_users_with_ranks.sort(
            by=["platform", "queries_rank"], descending=[False, False]
        )

        if not self.config.disable_write_usage:
            # Calculate write usage
            if self.config.use_server_side_aggregation:
                write_lf = self.load_write_usage_server_side_aggregation(datasets_lf)
            else:
                write_lf = self.load_write_usage(datasets_lf)
        else:
            logger.info("Write usage disabled")
            write_lf = polars.LazyFrame(
                schema={
                    "urn": polars.Categorical,
                    "platform": polars.Categorical,
                    "write_count": polars.Int64,
                }
            )

        usage_and_write_lf = (
            usage_with_top_users_with_ranks.join(
                write_lf, on="urn", how="full", suffix="_write"
            )
            .with_columns(polars.col("write_count").fill_null(polars.lit(0)))
            .with_columns(polars.col("totalSqlQueries").fill_null(polars.lit(0)))
        )

        # If we get a dataset from the operation aspect index only then we have to use its urn and platform
        usage_and_write_lf = usage_and_write_lf.with_columns(
            polars.col("urn").fill_null(polars.col("urn_write"))
        )
        usage_and_write_lf = usage_and_write_lf.with_columns(
            polars.col("platform").fill_null(polars.col("platform_write"))
        )

        usage_and_write_lf = self.gen_rank_and_percentile(
            usage_and_write_lf, "write_count", "urn", "platform", "write_"
        )
        return usage_and_write_lf

    def load_data_from_es_to_lf(
        self,
        index: str,
        schema: Dict,
        query: Dict,
        process_function: Callable,
        aggregation_key: Optional[str] = None,
        file_to_load: Optional[str] = None,
    ) -> polars.LazyFrame:
        data = self.load_data_from_es(
            index=index,
            query=query,
            process_function=process_function,
            aggregation_key=aggregation_key,
        )

        if not self.config.streaming_mode:
            return polars.LazyFrame(data, schema)
        else:
            assert self.temp_dir is not None, (
                "In Streaming mode temp dir should be set. Normally this should not happen..."
            )

            with tempfile.NamedTemporaryFile(
                delete=False,
                mode="wb",
                dir=self.temp_dir.name,
                prefix=f"{index}_",
                suffix=".parquet",
            ) as temp_file:
                tempfile_name = temp_file.name
                with pq.ParquetWriter(
                    tempfile_name, self.polars_to_arrow_schema(schema)
                ) as writer:
                    logger.debug(f"Creating temporary file {tempfile_name}")

                    self.batch_write_parquet(
                        data,
                        schema,
                        temp_file.name,
                        parquet_writer=writer,
                    )
                # Scan parquet fails in some cases with
            # thread 'polars-1' panicked at crates/polars-parquet/src/arrow/read/deserialize/dictionary_encoded/required_masked_dense.rs:113:72:
            # called `Option::unwrap()` on a `None` value
            # Which only happens if we don't collect immediately
            # return polars.scan_parquet(temp_file.name, schema=schema, low_memory=True).collect().lazy()
            return polars.scan_parquet(temp_file.name, schema=schema, low_memory=True)

    def load_dataset_usage(self) -> polars.LazyFrame:
        index = "dataset_datasetusagestatisticsaspect_v1"
        schema = {
            "timestampMillis": polars.Int64,
            "urn": polars.Categorical,
            "platform": polars.Categorical,
            "eventGranularity": polars.String,
            "totalSqlQueries": polars.Int64,
            "uniqueUserCount": polars.Int64,
            "userCounts": polars.List(
                polars.Struct(
                    {
                        "count": polars.Int64,
                        "user": polars.String,
                        "userEmail": polars.String,
                    }
                )
            ),
        }

        return self.load_data_from_es_to_lf(
            schema=schema,
            index=index,
            query=QueryBuilder.get_dataset_usage_query(self.config.lookback_days),
            process_function=self.process_batch,
        )

    def get_datasets(self) -> polars.LazyFrame:
        schema = {
            "entity_urn": polars.Categorical,
            "removed": polars.Boolean,
            "last_modified_at": polars.Int64,
            "siblings": polars.List(polars.String),
            "combinedSearchRankingMultiplier": polars.Float64,
            "isView": polars.Boolean,
        }

        return self.load_data_from_es_to_lf(
            schema=schema,
            index="datasetindex_v2",
            query=QueryBuilder.get_dataset_entities_query(),
            process_function=self.soft_deleted_batch,
        )

    def generate_top_users(
        self, lf: polars.LazyFrame, count_field_name: str = "count"
    ) -> polars.LazyFrame:
        #  Getting top users

        top_users = (
            lf.explode("userCounts")
            .unnest("userCounts")
            .filter(polars.col("user").is_not_null())
        )

        top_users = (
            top_users.group_by("urn", "platform", "user")
            .agg(polars.col(count_field_name).sum().alias("count"))
            .sort(by=["urn"], descending=[False])
        )
        top_users = top_users.with_columns(
            polars.col("user")
            .unique()
            .count()
            .over("platform", "urn")
            .alias("distinct_user")
        )

        top_users = top_users.with_columns(
            polars.col("count").sum().over("urn").alias("total_user_count")
        )

        top_users = top_users.with_columns(
            polars.col("count")
            .rank(descending=True, method="ordinal")
            .over("platform", "urn")
            .alias("user_rank")
        )

        top_users = top_users.filter(polars.col("user_rank") <= 10).sort(
            by=["urn", "user_rank"], descending=[False, False]
        )

        top_users = (
            top_users.group_by("urn", "platform", "distinct_user", "total_user_count")
            .agg(
                # polars.concat_list(polars.col("user")).alias("top_users")
                polars.when(polars.col("user").count() == 0)
                .then(polars.concat_list(polars.lit([])))
                .otherwise(polars.concat_list(polars.col("user")))
                .alias("top_users")
                # polars.when(polars.col("user").count() == 0).then(polars.concat_list(polars.lit(polars.Series([], dtype=polars.Int64)))).otherwise(polars.concat_list(polars.col("user"))).alias("top_users")
            )
            .select(
                ["urn", "platform", "top_users", "distinct_user", "total_user_count"]
            )
        )

        top_users = self.gen_rank_and_percentile(
            top_users, "distinct_user", "urn", "platform", "distinct_user_"
        )
        top_users = top_users.filter(polars.col("urn").is_not_null())
        return top_users

    def load_es_data(
        self,
        query: Dict,
        server: Union[OpenSearch, Elasticsearch],
        index: str,
        process_function: Callable,
        aggregation_key: Optional[str] = None,
        batch_size: int = 1000,
        delay: Optional[float] = None,
    ) -> Iterable[Dict[str, Any]]:
        processed_count = 0
        scroll_id = None
        while True:
            with PerfTimer() as timer:
                logger.debug(f"ES query: {query}")
                if not scroll_id:
                    logger.debug(
                        f"Getting inital data from index {index} without scroll id"
                    )
                    results = server.search(
                        body=query,
                        size=batch_size,
                        scroll="2m",
                        index=index,
                        params=(
                            {"timeout": self.config.query_timeout}
                            if self.config.search_index.opensearch_dialect
                            else {"request_timeout": self.config.query_timeout}
                        ),
                    )
                else:
                    logger.debug(
                        f"Getting data from index {index} using scroll_id: {scroll_id}"
                    )
                    results = server.scroll(
                        scroll_id=scroll_id,
                        scroll="2m",
                        params=(
                            {"timeout": self.config.query_timeout}
                            if self.config.search_index.opensearch_dialect
                            else {"request_timeout": self.config.query_timeout}
                        ),
                    )
                scroll_id = results["_scroll_id"]

                if not aggregation_key:
                    yield from process_function(results["hits"]["hits"])

                    time_taken = timer.elapsed_seconds()
                    processed_count += len(results["hits"]["hits"])
                    logger.info(
                        f"Processed {len(results['hits']['hits'])} data from {index} index in {time_taken:.3f} seconds. Total: {processed_count} processed."
                    )
                    if len(results["hits"]["hits"]) < batch_size:
                        break
                else:
                    yield from process_function(
                        results["aggregations"][aggregation_key]["buckets"]
                    )
                    if (
                        len((results["aggregations"][aggregation_key]["buckets"]))
                        < batch_size
                    ):
                        break
            if delay:
                logger.debug(
                    f"Sleeping for {delay} seconds before getting next batch from ES"
                )
                time.sleep(delay)

    def get_report(self) -> SourceReport:
        return self.report



from acryl_datahub_cloud.datahub_restore.do_restore import _parse_response


def test_parse_args() -> None:
    response = {
        "value": "{args=RestoreIndicesArgs(start=0, batchSize=100, limit=0, numThreads=1, batchDelayMs=1000, gePitEpochMs=0, lePitEpochMs=1738171914414, aspectName=null, aspectNames=[], urn=urn:li:container:00574c9e83ccb045d2d5039f4b1ede9a, urnLike=null, urnBasedPagination=false, lastUrn=, lastAspect=), result=RestoreIndicesResult(ignored=0, rowsMigrated=8, timeSqlQueryMs=2, timeGetRowMs=4, timeUrnMs=0, timeEntityRegistryCheckMs=0, aspectCheckMs=0, createRecordMs=0, sendMessageMs=37, defaultAspectsCreated=0, lastUrn=urn:li:container:00574c9e83ccb045d2d5039f4b1ede9a, lastAspect=testResults)}",
    }
    assert _parse_response(response) == {
        "args": {
            "start": 0,
            "aspectName": None,
            "aspectNames": [],
            "batchDelayMs": 1000,
            "batchSize": 100,
            "gePitEpochMs": 0,
            "lastAspect": "",
            "lastUrn": "",
            "lePitEpochMs": 1738171914414,
            "limit": 0,
            "numThreads": 1,
            "urn": "urn:li:container:00574c9e83ccb045d2d5039f4b1ede9a",
            "urnBasedPagination": False,
            "urnLike": None,
        },
        "result": {
            "aspectCheckMs": 0,
            "createRecordMs": 0,
            "defaultAspectsCreated": 0,
            "ignored": 0,
            "lastAspect": "testResults",
            "lastUrn": "urn:li:container:00574c9e83ccb045d2d5039f4b1ede9a",
            "rowsMigrated": 8,
            "sendMessageMs": 37,
            "timeEntityRegistryCheckMs": 0,
            "timeGetRowMs": 4,
            "timeSqlQueryMs": 2,
            "timeUrnMs": 0,
        },
    }



"""
This module contains the classes that represent assertions. These
classes are used to provide a user-friendly interface for creating and
managing assertions.

The actual Assertion Entity classes are defined in `metadata-ingestion/src/datahub/sdk`.
"""

import logging
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Optional, Union

from typing_extensions import Self

from acryl_datahub_cloud.sdk.assertion_input import (
    ASSERTION_MONITOR_DEFAULT_TRAINING_LOOKBACK_WINDOW_DAYS,
    DEFAULT_DETECTION_MECHANISM,
    DEFAULT_SCHEDULE,
    DEFAULT_SENSITIVITY,
    AssertionIncidentBehavior,
    DetectionMechanism,
    ExclusionWindowTypes,
    FixedRangeExclusionWindow,
    InferenceSensitivity,
    _DetectionMechanismTypes,
)
from acryl_datahub_cloud.sdk.entities.assertion import Assertion
from acryl_datahub_cloud.sdk.entities.monitor import (
    Monitor,
    _get_nested_field_for_entity_with_default,
)
from acryl_datahub_cloud.sdk.errors import SDKNotYetSupportedError
from datahub.emitter.mce_builder import parse_ts_millis
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn, TagUrn

logger = logging.getLogger(__name__)


class AssertionMode(Enum):
    """
    The mode of an assertion, e.g. whether it is active or inactive.
    """

    # Note: Modeled here after MonitorStatus but called AssertionMode in this user facing interface
    # to keep all naming related to assertions.
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    # PASSIVE = "PASSIVE" # Not supported in the user facing interface.


class _HasSchedule:
    """
    Mixin class that provides schedule functionality for assertions.
    """

    def __init__(self, schedule: models.CronScheduleClass) -> None:
        self._schedule = schedule

    @property
    def schedule(self) -> models.CronScheduleClass:
        return self._schedule

    @staticmethod
    def _get_schedule(monitor: Monitor) -> models.CronScheduleClass:
        """Get the schedule from the monitor."""
        assertion_evaluation_specs = _get_nested_field_for_entity_with_default(
            monitor,
            "info.assertionMonitor.assertions",
            [],
        )
        if len(assertion_evaluation_specs) == 0:
            return DEFAULT_SCHEDULE
        assertion_evaluation_spec = assertion_evaluation_specs[0]
        schedule = assertion_evaluation_spec.schedule
        if schedule is None:
            return DEFAULT_SCHEDULE
        return schedule


class _HasSmartFunctionality:
    """
    Mixin class that provides smart functionality for assertions.
    """

    _SUPPORTED_WITH_FILTER_ASSERTION_TYPES = (
        models.FreshnessAssertionInfoClass,
        models.VolumeAssertionInfoClass,
    )

    def __init__(
        self,
        *,
        sensitivity: InferenceSensitivity = DEFAULT_SENSITIVITY,
        exclusion_windows: list[ExclusionWindowTypes],
        training_data_lookback_days: int = ASSERTION_MONITOR_DEFAULT_TRAINING_LOOKBACK_WINDOW_DAYS,
        incident_behavior: list[AssertionIncidentBehavior],
        detection_mechanism: Optional[
            _DetectionMechanismTypes
        ] = DEFAULT_DETECTION_MECHANISM,
    ) -> None:
        """
        Initialize the smart functionality mixin.

        Args:
            sensitivity: The sensitivity of the assertion (low, medium, high).
            exclusion_windows: The exclusion windows of the assertion.
            training_data_lookback_days: The max number of days of data to use for training the assertion.
            incident_behavior: Whether to raise or resolve an incident when the assertion fails / passes.
            detection_mechanism: The detection mechanism of the assertion.
            **kwargs: Additional arguments to pass to the parent class (_Assertion).
        """
        self._sensitivity = sensitivity
        self._exclusion_windows = exclusion_windows
        self._training_data_lookback_days = training_data_lookback_days
        self._incident_behavior = incident_behavior
        self._detection_mechanism = detection_mechanism

    @property
    def sensitivity(self) -> InferenceSensitivity:
        return self._sensitivity

    @property
    def exclusion_windows(self) -> list[ExclusionWindowTypes]:
        return self._exclusion_windows

    @property
    def training_data_lookback_days(self) -> int:
        return self._training_data_lookback_days

    @property
    def incident_behavior(self) -> list[AssertionIncidentBehavior]:
        return self._incident_behavior

    @property
    def detection_mechanism(self) -> Optional[_DetectionMechanismTypes]:
        return self._detection_mechanism

    @staticmethod
    def _get_sensitivity(monitor: Monitor) -> InferenceSensitivity:
        # 1. Check if the monitor has a sensitivity field
        raw_sensitivity = _get_nested_field_for_entity_with_default(
            monitor,
            "info.assertionMonitor.settings.adjustmentSettings.sensitivity.level",
            DEFAULT_SENSITIVITY,
        )

        # 2. Convert the raw sensitivity to the SDK sensitivity enum (1-3: LOW, 4-6: MEDIUM, 7-10: HIGH)
        return InferenceSensitivity.parse(raw_sensitivity)

    @staticmethod
    def _get_exclusion_windows(monitor: Monitor) -> list[ExclusionWindowTypes]:
        # 1. Check if the monitor has an exclusion windows field
        raw_windows = monitor.exclusion_windows or []

        # 2. Convert the raw exclusion windows to the SDK exclusion windows
        exclusion_windows = []
        for raw_window in raw_windows:
            if raw_window.type == models.AssertionExclusionWindowTypeClass.FIXED_RANGE:
                if raw_window.fixedRange is None:
                    logger.warning(
                        f"Monitor {monitor.urn} has a fixed range exclusion window with no fixed range, skipping"
                    )
                    continue
                exclusion_windows.append(
                    FixedRangeExclusionWindow(
                        start=parse_ts_millis(raw_window.fixedRange.startTimeMillis),
                        end=parse_ts_millis(raw_window.fixedRange.endTimeMillis),
                    )
                )
            else:
                raise SDKNotYetSupportedError(
                    f"AssertionExclusionWindowType {raw_window.type}"
                )
        return exclusion_windows

    @staticmethod
    def _get_training_data_lookback_days(monitor: Monitor) -> int:
        retrieved = monitor.training_data_lookback_days
        if (
            retrieved is None
        ):  # Explicitly check for None since retrieved can be 0 which is falsy
            return ASSERTION_MONITOR_DEFAULT_TRAINING_LOOKBACK_WINDOW_DAYS
        assert isinstance(retrieved, int)
        return retrieved

    @staticmethod
    def _get_detection_mechanism(
        assertion: Assertion,
        monitor: Monitor,
        default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
    ) -> Optional[_DetectionMechanismTypes]:
        """Get the detection mechanism from the monitor and assertion."""
        if not _HasSmartFunctionality._has_valid_monitor_info(monitor):
            return default

        # 1. Check if the assertion has a parameters field
        def _warn_and_return_default_detection_mechanism(
            field_name: str,
            default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
        ) -> Optional[_DetectionMechanismTypes]:
            logger.warning(
                f"Monitor {monitor.urn} does not have an `{field_name}` field, defaulting detection mechanism to {default}"
            )
            return default

        parameters = _HasSmartFunctionality._get_assertion_parameters(monitor, default)
        if parameters is None:
            return _warn_and_return_default_detection_mechanism("parameters", default)

        # 2. Convert the raw detection mechanism to the SDK detection mechanism
        if parameters.type in [
            models.AssertionEvaluationParametersTypeClass.DATASET_FRESHNESS,
            models.AssertionEvaluationParametersTypeClass.DATASET_VOLUME,
        ]:
            if assertion.info is None:
                return _warn_and_return_default_detection_mechanism("info", default)
            if isinstance(assertion.info, models.VolumeAssertionInfoClass):
                return _HasSmartFunctionality._get_volume_detection_mechanism(
                    assertion, parameters, default
                )
            elif isinstance(assertion.info, models.FreshnessAssertionInfoClass):
                return _HasSmartFunctionality._get_freshness_detection_mechanism(
                    assertion, parameters, default
                )
            # TODO: Consider moving the detection mechanism logic to the assertion classes themselves e.g. _get_assertion_specific_detection_mechanism as an abstract method
            # TODO: Add support here for other detection mechanisms when other assertion types are supported
            else:
                raise SDKNotYetSupportedError(
                    f"AssertionType {type(assertion.info).__name__}"
                )
        else:
            raise SDKNotYetSupportedError(
                f"AssertionEvaluationParametersType {parameters.type} not supported"
            )

    @staticmethod
    def _has_valid_monitor_info(monitor: Monitor) -> bool:
        """Check if monitor has valid info and assertion monitor."""

        def _warn_and_return_false(field_name: str) -> bool:
            logger.warning(
                f"Monitor {monitor.urn} does not have an `{field_name}` field, defaulting detection mechanism to {DEFAULT_DETECTION_MECHANISM}"
            )
            return False

        if monitor.info is None:
            return _warn_and_return_false("info")
        if monitor.info.assertionMonitor is None:
            return _warn_and_return_false("assertionMonitor")
        if (
            monitor.info.assertionMonitor.assertions is None
            or len(monitor.info.assertionMonitor.assertions) == 0
        ):
            return _warn_and_return_false("assertionMonitor.assertions")

        return True

    @staticmethod
    def _get_assertion_parameters(
        monitor: Monitor,
        default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
    ) -> Optional[models.AssertionEvaluationParametersClass]:
        """Get the assertion parameters from the monitor."""
        # We know these are not None from _has_valid_monitor_info check
        assert (
            monitor is not None
            and monitor.info is not None
            and monitor.info.assertionMonitor is not None
        )
        assertion_monitor = monitor.info.assertionMonitor
        assert (
            assertion_monitor is not None and assertion_monitor.assertions is not None
        )
        assertions = assertion_monitor.assertions

        if assertions[0].parameters is None:
            logger.warning(
                f"Monitor {monitor.urn} does not have a assertionMonitor.assertions[0].parameters, defaulting detection mechanism to {default}"
            )
            return None
        return assertions[0].parameters

    @staticmethod
    def _get_freshness_detection_mechanism(
        assertion: Assertion,
        parameters: models.AssertionEvaluationParametersClass,
        default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
    ) -> Optional[_DetectionMechanismTypes]:
        """Get the detection mechanism for freshness assertions."""
        if parameters.datasetFreshnessParameters is None:
            logger.warning(
                f"Monitor does not have datasetFreshnessParameters, defaulting detection mechanism to {DEFAULT_DETECTION_MECHANISM}"
            )
            return default

        source_type = parameters.datasetFreshnessParameters.sourceType
        if source_type == models.DatasetFreshnessSourceTypeClass.INFORMATION_SCHEMA:
            return DetectionMechanism.INFORMATION_SCHEMA
        elif source_type == models.DatasetFreshnessSourceTypeClass.AUDIT_LOG:
            return DetectionMechanism.AUDIT_LOG
        elif source_type == models.DatasetFreshnessSourceTypeClass.FIELD_VALUE:
            return _HasSmartFunctionality._get_field_value_detection_mechanism(
                assertion, parameters
            )
        elif source_type == models.DatasetFreshnessSourceTypeClass.DATAHUB_OPERATION:
            return DetectionMechanism.DATAHUB_OPERATION
        elif source_type == models.DatasetFreshnessSourceTypeClass.FILE_METADATA:
            raise SDKNotYetSupportedError("FILE_METADATA DatasetFreshnessSourceType")
        else:
            raise SDKNotYetSupportedError(f"DatasetFreshnessSourceType {source_type}")

    @staticmethod
    def _get_volume_detection_mechanism(
        assertion: Assertion,
        parameters: models.AssertionEvaluationParametersClass,
        default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
    ) -> _DetectionMechanismTypes:
        """Get the detection mechanism for volume assertions."""
        if parameters.datasetVolumeParameters is None:
            logger.warning(
                f"Monitor does not have datasetVolumeParameters, defaulting detection mechanism to {DEFAULT_DETECTION_MECHANISM}"
            )
            if default is None:
                return DEFAULT_DETECTION_MECHANISM
            else:
                return default

        source_type = parameters.datasetVolumeParameters.sourceType
        if source_type == models.DatasetVolumeSourceTypeClass.INFORMATION_SCHEMA:
            return DetectionMechanism.INFORMATION_SCHEMA
        elif source_type == models.DatasetVolumeSourceTypeClass.QUERY:
            additional_filter = _HasSmartFunctionality._get_additional_filter(assertion)
            return DetectionMechanism.QUERY(additional_filter=additional_filter)
        elif source_type == models.DatasetVolumeSourceTypeClass.DATAHUB_DATASET_PROFILE:
            return DetectionMechanism.DATASET_PROFILE
        else:
            raise SDKNotYetSupportedError(f"DatasetVolumeSourceType {source_type}")

    @staticmethod
    def _get_field_value_detection_mechanism(
        assertion: Assertion,
        parameters: models.AssertionEvaluationParametersClass,
    ) -> _DetectionMechanismTypes:
        """Get the detection mechanism for field value based freshness."""
        # We know datasetFreshnessParameters is not None from _get_freshness_detection_mechanism check
        assert parameters.datasetFreshnessParameters is not None
        field = parameters.datasetFreshnessParameters.field

        if field is None or field.kind is None:
            logger.warning(
                f"Monitor does not have valid field info, defaulting detection mechanism to {DEFAULT_DETECTION_MECHANISM}"
            )
            return DEFAULT_DETECTION_MECHANISM

        column_name = field.path
        additional_filter = _HasSmartFunctionality._get_additional_filter(assertion)

        if field.kind == models.FreshnessFieldKindClass.LAST_MODIFIED:
            return DetectionMechanism.LAST_MODIFIED_COLUMN(
                column_name=column_name, additional_filter=additional_filter
            )
        elif field.kind == models.FreshnessFieldKindClass.HIGH_WATERMARK:
            return DetectionMechanism.HIGH_WATERMARK_COLUMN(
                column_name=column_name, additional_filter=additional_filter
            )
        else:
            raise SDKNotYetSupportedError(f"FreshnessFieldKind {field.kind}")

    @staticmethod
    def _get_additional_filter(assertion: Assertion) -> Optional[str]:
        """Get the additional filter SQL from the assertion."""
        if assertion.info is None:
            logger.warning(
                f"Assertion {assertion.urn} does not have an info, defaulting additional filter to None"
            )
            return None
        if (
            not isinstance(
                assertion.info,
                _HasSmartFunctionality._SUPPORTED_WITH_FILTER_ASSERTION_TYPES,
            )
            or assertion.info.filter is None
        ):
            logger.warning(
                f"Assertion {assertion.urn} does not have a filter, defaulting additional filter to None"
            )
            return None
        if assertion.info.filter.type != models.DatasetFilterTypeClass.SQL:
            raise SDKNotYetSupportedError(
                f"DatasetFilterType {assertion.info.filter.type}"
            )
        return assertion.info.filter.sql


class _AssertionPublic(ABC):
    """
    Abstract base class that represents a public facing assertion and contains the common properties of all assertions.
    """

    def __init__(
        self,
        *,
        urn: AssertionUrn,
        dataset_urn: DatasetUrn,
        display_name: str,
        mode: AssertionMode,
        tags: list[TagUrn],
        created_by: Optional[CorpUserUrn] = None,
        created_at: Union[datetime, None] = None,
        updated_by: Optional[CorpUserUrn] = None,
        updated_at: Optional[datetime] = None,
    ):
        """
        Initialize the public facing assertion class.

        Args:
            urn: The urn of the assertion.
            dataset_urn: The urn of the dataset that the assertion is for.
            display_name: The display name of the assertion.
            mode: The mode of the assertion (active, inactive).
            tags: The tags of the assertion.
            created_by: The urn of the user that created the assertion.
            created_at: The timestamp of when the assertion was created.
            updated_by: The urn of the user that updated the assertion.
            updated_at: The timestamp of when the assertion was updated.
        """
        self._urn = urn
        self._dataset_urn = dataset_urn
        self._display_name = display_name
        self._mode = mode
        self._created_by = created_by
        self._created_at = created_at
        self._updated_by = updated_by
        self._updated_at = updated_at
        self._tags = tags

    @property
    def urn(self) -> AssertionUrn:
        return self._urn

    @property
    def dataset_urn(self) -> DatasetUrn:
        return self._dataset_urn

    @property
    def display_name(self) -> str:
        return self._display_name

    @property
    def mode(self) -> AssertionMode:
        return self._mode

    @property
    def created_by(self) -> Optional[CorpUserUrn]:
        return self._created_by

    @property
    def created_at(self) -> Union[datetime, None]:
        return self._created_at

    @property
    def updated_by(self) -> Optional[CorpUserUrn]:
        return self._updated_by

    @property
    def updated_at(self) -> Union[datetime, None]:
        return self._updated_at

    @property
    def tags(self) -> list[TagUrn]:
        return self._tags

    @staticmethod
    def _get_incident_behavior(assertion: Assertion) -> list[AssertionIncidentBehavior]:
        incident_behaviors = []
        for action in assertion.on_failure + assertion.on_success:
            if action.type == models.AssertionActionTypeClass.RAISE_INCIDENT:
                incident_behaviors.append(AssertionIncidentBehavior.RAISE_ON_FAIL)
            elif action.type == models.AssertionActionTypeClass.RESOLVE_INCIDENT:
                incident_behaviors.append(AssertionIncidentBehavior.RESOLVE_ON_PASS)

        return incident_behaviors

    @staticmethod
    def _get_created_by(assertion: Assertion) -> Optional[CorpUserUrn]:
        if assertion.source is None:
            logger.warning(f"Assertion {assertion.urn} does not have a source")
            return None
        if isinstance(assertion.source, models.AssertionSourceClass):
            if assertion.source.created is None:
                logger.warning(
                    f"Assertion {assertion.urn} does not have a created by in the source"
                )
                return None
            return CorpUserUrn.from_string(assertion.source.created.actor)
        elif isinstance(assertion.source, models.AssertionSourceTypeClass):
            logger.warning(
                f"Assertion {assertion.urn} has a source type with no created by"
            )
            return None
        return None

    @staticmethod
    def _get_created_at(assertion: Assertion) -> Union[datetime, None]:
        if assertion.source is None:
            logger.warning(f"Assertion {assertion.urn} does not have a source")
            return None
        if isinstance(assertion.source, models.AssertionSourceClass):
            if assertion.source.created is None:
                logger.warning(
                    f"Assertion {assertion.urn} does not have a created by in the source"
                )
                return None
            return parse_ts_millis(assertion.source.created.time)
        elif isinstance(assertion.source, models.AssertionSourceTypeClass):
            logger.warning(
                f"Assertion {assertion.urn} has a source type with no created by"
            )
            return None
        return None

    @staticmethod
    def _get_updated_by(assertion: Assertion) -> Optional[CorpUserUrn]:
        if assertion.last_updated is None:
            logger.warning(f"Assertion {assertion.urn} does not have a last updated")
            return None
        return CorpUserUrn.from_string(assertion.last_updated.actor)

    @staticmethod
    def _get_updated_at(assertion: Assertion) -> Union[datetime, None]:
        if assertion.last_updated is None:
            logger.warning(f"Assertion {assertion.urn} does not have a last updated")
            return None
        return parse_ts_millis(assertion.last_updated.time)

    @staticmethod
    def _get_tags(assertion: Assertion) -> list[TagUrn]:
        return [TagUrn.from_string(t.tag) for t in assertion.tags or []]

    @staticmethod
    def _get_mode(monitor: Monitor) -> AssertionMode:
        if monitor.info is None:
            logger.warning(
                f"Monitor {monitor.urn} does not have a info, defaulting status to INACTIVE"
            )
            return AssertionMode.INACTIVE
        return AssertionMode(monitor.info.status.mode)

    @classmethod
    @abstractmethod
    def _from_entities(
        cls,
        assertion: Assertion,
        monitor: Monitor,
    ) -> Self:
        """
        Create an assertion from the assertion and monitor entities.

        Note: This is a private method since it is intended to be called internally by the client.
        """
        pass


class SmartFreshnessAssertion(_HasSchedule, _HasSmartFunctionality, _AssertionPublic):
    """
    A class that represents a smart freshness assertion.
    """

    def __init__(
        self,
        *,
        urn: AssertionUrn,
        dataset_urn: DatasetUrn,
        display_name: str,
        mode: AssertionMode,
        schedule: models.CronScheduleClass = DEFAULT_SCHEDULE,
        sensitivity: InferenceSensitivity = DEFAULT_SENSITIVITY,
        exclusion_windows: list[ExclusionWindowTypes],
        training_data_lookback_days: int = ASSERTION_MONITOR_DEFAULT_TRAINING_LOOKBACK_WINDOW_DAYS,
        incident_behavior: list[AssertionIncidentBehavior],
        detection_mechanism: Optional[
            _DetectionMechanismTypes
        ] = DEFAULT_DETECTION_MECHANISM,
        tags: list[TagUrn],
        created_by: Optional[CorpUserUrn] = None,
        created_at: Union[datetime, None] = None,
        updated_by: Optional[CorpUserUrn] = None,
        updated_at: Optional[datetime] = None,
    ):
        """
        Initialize a smart freshness assertion.

        Note: Values can be accessed, but not set on the assertion object.
        To update an assertion, use the `upsert_*` method.
        Args:
            urn: The urn of the assertion.
            dataset_urn: The urn of the dataset that the assertion is for.
            display_name: The display name of the assertion.
            mode: The mode of the assertion (active, inactive).
            schedule: The schedule of the assertion.
            sensitivity: The sensitivity of the assertion (low, medium, high).
            exclusion_windows: The exclusion windows of the assertion.
            training_data_lookback_days: The max number of days of data to use for training the assertion.
            incident_behavior: Whether to raise or resolve an incident when the assertion fails / passes.
            detection_mechanism: The detection mechanism of the assertion.
            tags: The tags applied to the assertion.
            created_by: The urn of the user that created the assertion.
            created_at: The timestamp of when the assertion was created.
            updated_by: The urn of the user that updated the assertion.
            updated_at: The timestamp of when the assertion was updated.
        """
        # Initialize the mixins first
        _HasSchedule.__init__(self, schedule=schedule)
        _HasSmartFunctionality.__init__(
            self,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            detection_mechanism=detection_mechanism,
        )
        # Then initialize the parent class
        _AssertionPublic.__init__(
            self,
            urn=urn,
            dataset_urn=dataset_urn,
            display_name=display_name,
            mode=mode,
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
            tags=tags,
        )

    @classmethod
    def _from_entities(cls, assertion: Assertion, monitor: Monitor) -> Self:
        """
        Create a smart freshness assertion from the assertion and monitor entities.

        Note: This is a private method since it is intended to be called internally by the client.
        """
        return cls(
            urn=assertion.urn,
            dataset_urn=assertion.dataset,
            display_name=assertion.description or "",
            mode=cls._get_mode(monitor),
            schedule=cls._get_schedule(monitor),
            sensitivity=cls._get_sensitivity(monitor),
            exclusion_windows=cls._get_exclusion_windows(monitor),
            training_data_lookback_days=cls._get_training_data_lookback_days(monitor),
            incident_behavior=cls._get_incident_behavior(assertion),
            detection_mechanism=cls._get_detection_mechanism(assertion, monitor),
            created_by=cls._get_created_by(assertion),
            created_at=cls._get_created_at(assertion),
            updated_by=cls._get_updated_by(assertion),
            updated_at=cls._get_updated_at(assertion),
            tags=cls._get_tags(assertion),
        )


class SmartVolumeAssertion(_HasSchedule, _HasSmartFunctionality, _AssertionPublic):
    """
    A class that represents a smart volume assertion.
    """

    def __init__(
        self,
        *,
        urn: AssertionUrn,
        dataset_urn: DatasetUrn,
        display_name: str,
        mode: AssertionMode,
        schedule: models.CronScheduleClass,
        sensitivity: InferenceSensitivity = DEFAULT_SENSITIVITY,
        exclusion_windows: list[ExclusionWindowTypes],
        training_data_lookback_days: int = ASSERTION_MONITOR_DEFAULT_TRAINING_LOOKBACK_WINDOW_DAYS,
        incident_behavior: list[AssertionIncidentBehavior],
        detection_mechanism: Optional[
            _DetectionMechanismTypes
        ] = DEFAULT_DETECTION_MECHANISM,
        tags: list[TagUrn],
        created_by: Optional[CorpUserUrn] = None,
        created_at: Union[datetime, None] = None,
        updated_by: Optional[CorpUserUrn] = None,
        updated_at: Optional[datetime] = None,
    ):
        """
        Initialize a smart volume assertion.

        Note: Values can be accessed, but not set on the assertion object.
        To update an assertion, use the `upsert_*` method.
        Args:
            urn: The urn of the assertion.
            dataset_urn: The urn of the dataset that the assertion is for.
            display_name: The display name of the assertion.
            mode: The mode of the assertion (active, inactive).
            schedule: The schedule of the assertion.
            sensitivity: The sensitivity of the assertion (low, medium, high).
            exclusion_windows: The exclusion windows of the assertion.
            training_data_lookback_days: The max number of days of data to use for training the assertion.
            incident_behavior: Whether to raise or resolve an incident when the assertion fails / passes.
            detection_mechanism: The detection mechanism of the assertion.
            tags: The tags applied to the assertion.
            created_by: The urn of the user that created the assertion.
            created_at: The timestamp of when the assertion was created.
            updated_by: The urn of the user that updated the assertion.
            updated_at: The timestamp of when the assertion was updated.
        """
        # Initialize the mixins first
        _HasSchedule.__init__(self, schedule=schedule)
        _HasSmartFunctionality.__init__(
            self,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            detection_mechanism=detection_mechanism,
        )
        # Then initialize the parent class
        _AssertionPublic.__init__(
            self,
            urn=urn,
            dataset_urn=dataset_urn,
            display_name=display_name,
            mode=mode,
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
            tags=tags,
        )

    @classmethod
    def _from_entities(cls, assertion: Assertion, monitor: Monitor) -> Self:
        """
        Create a smart freshness assertion from the assertion and monitor entities.

        Note: This is a private method since it is intended to be called internally by the client.
        """
        return cls(
            urn=assertion.urn,
            dataset_urn=assertion.dataset,
            display_name=assertion.description or "",
            mode=cls._get_mode(monitor),
            schedule=cls._get_schedule(monitor),
            sensitivity=cls._get_sensitivity(monitor),
            exclusion_windows=cls._get_exclusion_windows(monitor),
            training_data_lookback_days=cls._get_training_data_lookback_days(monitor),
            incident_behavior=cls._get_incident_behavior(assertion),
            detection_mechanism=cls._get_detection_mechanism(assertion, monitor),
            created_by=cls._get_created_by(assertion),
            created_at=cls._get_created_at(assertion),
            updated_by=cls._get_updated_by(assertion),
            updated_at=cls._get_updated_at(assertion),
            tags=cls._get_tags(assertion),
        )


AssertionTypes = Union[
    SmartFreshnessAssertion,
    SmartVolumeAssertion,
    # TODO: Add other assertion types here as we add them.
]
